import os
from abc import ABC, abstractmethod
from copy import copy
from http.cookies import SimpleCookie
from typing import AsyncIterable  # noqa
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    AsyncIterator,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseConfig, BaseModel, FilePath, validator  # noqa
from pydantic.generics import GenericModel  # noqa
from starlette.datastructures import URL as URL  # noqa: F401
from starlette.datastructures import Address as Address  # noqa: F401
from starlette.datastructures import FormData as FormData  # noqa: F401
from starlette.datastructures import Headers as Headers  # noqa: F401
from starlette.datastructures import MutableHeaders as MutableHeaders  # noqa
from starlette.datastructures import QueryParams as QueryParams  # noqa: F401
from starlette.datastructures import State as StarletteStateClass  # noqa: F401
from starlette.datastructures import UploadFile as UploadFile  # noqa
from starlette.datastructures import URLPath as URLPath  # noqa: F401
from starlette.responses import FileResponse, RedirectResponse
from starlette.responses import Response as StarletteResponse
from starlette.responses import StreamingResponse
from typing_extensions import Literal, ParamSpec

from esmerald.backgound import BackgroundTask, BackgroundTasks  # noqa
from esmerald.responses import TemplateResponse

P = ParamSpec("P")
R = TypeVar("R", bound=StarletteResponse)

if TYPE_CHECKING:
    from esmerald.applications import Esmerald
    from esmerald.enums import MediaType


class Secret:
    def __init__(self, value: str):
        self._value = value

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}('**********')"

    def __str__(self) -> str:
        return self._value

    def __bool__(self) -> bool:
        return bool(self._value)

    def __len__(self) -> int:
        return len(self._value)


class State(StarletteStateClass):

    state: Dict[str, Any]

    def __copy__(self) -> "State":
        return self.__class__(copy(self._state))

    def __len__(self):
        return len(self._state)

    def __getattr__(self, key: str) -> Any:
        try:
            return self._state[key]
        except KeyError as e:
            raise AttributeError(f"State has no key '{key}'") from e

    def __getitem__(self, key: str) -> Any:
        return self._state[key]

    def copy(self) -> "State":
        return copy(self)


class Cookie(BaseModel):
    key: str
    value: Optional[str] = None
    max_age: Optional[int] = None
    expires: Optional[int] = None
    path: str = "/"
    domain: Optional[str] = None
    secure: Optional[bool] = None
    httponly: Optional[bool] = None
    samesite: Literal["lax", "strict", "none"] = "lax"
    description: Optional[str] = None

    def to_header(self, **kwargs: Any) -> str:
        simple_cookie: SimpleCookie = SimpleCookie()
        simple_cookie[self.key] = self.value or ""
        if self.max_age:
            simple_cookie[self.key]["max-age"] = self.max_age
        cookie_dict = self.dict()
        for key in ["expires", "path", "domain", "secure", "httponly", "samesite"]:
            if cookie_dict[key] is not None:
                simple_cookie[self.key][key] = cookie_dict[key]
        return simple_cookie.output(**kwargs).strip()


class ResponseContainer(GenericModel, ABC, Generic[R]):
    class Config(BaseConfig):
        arbitrary_types_allowed = True

    background: Optional[Union[BackgroundTask, BackgroundTasks]] = None
    headers: Dict[str, Any] = {}
    cookies: List[Cookie] = []

    @abstractmethod
    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Esmerald",
    ) -> R:  # pragma: no cover
        raise NotImplementedError("not implemented")


class File(ResponseContainer[FileResponse]):
    path: FilePath
    filename: str
    stat_result: Optional[os.stat_result] = None

    @validator("stat_result", always=True)
    def validate_status_code(  # pylint: disable=no-self-argument
        cls, value: Optional[os.stat_result], values: Dict[str, Any]
    ) -> os.stat_result:
        """Set the stat_result value for the given filepath."""
        return value or os.stat(cast("str", values.get("path")))

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Esmerald",
    ) -> FileResponse:
        return FileResponse(
            background=self.background,
            filename=self.filename,
            headers=headers,
            media_type=media_type,
            path=self.path,
            stat_result=self.stat_result,
            status_code=status_code,
        )


class Redirect(ResponseContainer[RedirectResponse]):
    path: str

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Esmerald",
    ) -> RedirectResponse:
        return RedirectResponse(
            headers=headers,
            status_code=status_code,
            url=self.path,
            background=self.background,
        )


class Stream(ResponseContainer[StreamingResponse]):

    iterator: Union[
        Iterator[Union[str, bytes]],
        Generator[Union[str, bytes], Any, Any],
        AsyncIterator[Union[str, bytes]],
        AsyncGenerator[Union[str, bytes], Any],
        Type[Iterator[Union[str, bytes]]],
        Type[AsyncIterator[Union[str, bytes]]],
        Callable[[], AsyncGenerator[Union[str, bytes], Any]],
        Callable[[], Generator[Union[str, bytes], Any, Any]],
    ]

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Esmerald",
    ) -> StreamingResponse:

        return StreamingResponse(
            background=self.background,
            content=self.iterator
            if isinstance(self.iterator, (Iterable, AsyncIterable))
            else self.iterator(),
            headers=headers,
            media_type=media_type,
            status_code=status_code,
        )


class Template(ResponseContainer[TemplateResponse]):
    name: str
    context: Optional[Dict[str, Any]] = None

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Esmerald",
    ) -> "TemplateResponse":
        from esmerald.exceptions import ImproperlyConfigured
        from esmerald.responses import TemplateResponse

        if not app.template_engine:
            raise ImproperlyConfigured("Template engine is not configured")
        return TemplateResponse(
            background=self.background,
            context=self.context,
            headers=headers,
            status_code=status_code,
            template_engine=app.template_engine,
            template_name=self.name,
            media_type=media_type,
        )


class ResponseHeader(BaseModel):
    value: Any = None

    @validator("value", always=True)
    def validate_value(
        cls, value: Any, values: Dict[str, Any]
    ) -> Any:  # pylint: disable=no-self-argument
        if value is not None:
            return value



from typing import Any, Dict

from esmerald.core.management.templates import TemplateDirective
from esmerald.utils.crypto import get_random_secret_key

SECRET_KEY_INSECURE_PREFIX = "esmerald-insecure-"


class Directive(TemplateDirective):
    help = (
        "Creates an Esmerald project directory structure for the given project "
        "name in the current directory or optionally in the given directory."
    )
    missing_args_message = "You must provide a project name."

    def handle(self, **options: Dict[str, Any]):
        project_name = options.pop("name")

        options["secret_key"] = SECRET_KEY_INSECURE_PREFIX + get_random_secret_key()
        super().handle("project", project_name, **options)



import inspect
from copy import copy
from enum import Enum, IntEnum
from inspect import Signature
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Awaitable,
    Callable,
    Dict,
    List,
    NoReturn,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from starlette import status
from starlette.middleware import Middleware as StarletteMiddleware
from starlette.requests import HTTPConnection
from starlette.responses import JSONResponse
from starlette.responses import Response as StarletteResponse
from starlette.routing import BaseRoute as StarletteBaseRoute
from starlette.routing import Host, Mount, NoMatchFound
from starlette.routing import Route as StarletteRoute
from starlette.routing import Router as StarletteRouter
from starlette.routing import WebSocketRoute as StarletteWebSocketRoute
from starlette.routing import compile_path
from starlette.types import ASGIApp, Receive, Scope, Send

from esmerald.conf import settings
from esmerald.datastructures import File, Redirect, URLPath
from esmerald.enums import HttpMethod, MediaType
from esmerald.exceptions import (
    ImproperlyConfigured,
    MethodNotAllowed,
    NotFound,
    ValidationErrorException,
)
from esmerald.injector import Inject
from esmerald.interceptors.types import Interceptor
from esmerald.openapi.datastructures import ResponseSpecification
from esmerald.requests import Request
from esmerald.responses import Response
from esmerald.routing.base import BaseHandlerMixin
from esmerald.routing.gateways import Gateway, WebSocketGateway
from esmerald.routing.views import APIView
from esmerald.transformers.datastructures import EsmeraldSignature as SignatureModel
from esmerald.transformers.model import TransformerModel
from esmerald.transformers.utils import get_signature
from esmerald.typing import Void, VoidType
from esmerald.urls import include
from esmerald.utils.constants import DATA, REDIRECT_STATUS_CODES, REQUEST, SOCKET
from esmerald.utils.helpers import is_async_callable, is_class_and_subclass
from esmerald.utils.url import clean_path
from esmerald.websockets import WebSocket, WebSocketClose

if TYPE_CHECKING:
    from openapi_schemas_pydantic.v3_1_0 import SecurityRequirement
    from pydantic.typing import AnyCallable

    from esmerald.applications import Esmerald
    from esmerald.exceptions import HTTPException
    from esmerald.permissions.types import Permission
    from esmerald.types import (
        APIGateHandler,
        AsyncAnyCallable,
        BackgroundTaskType,
        CoroutineHandler,
        Dependencies,
        ExceptionHandler,
        ExceptionHandlers,
        HTTPMethod,
        LifeSpanHandler,
        Middleware,
        ParentType,
        ResponseCookies,
        ResponseHeaders,
        ResponseType,
        RouteParent,
    )


class Parent:
    def create_signature_models(self, route: "RouteParent"):
        """
        Creates the signature models for the given routes.

        Args:
            route: The route for the signature model to be created.
        """
        if isinstance(route, (Include, Host)):
            for _route in route.routes:
                self.create_signature_models(_route)

        if isinstance(route, Gateway):
            if not route.handler.parent:
                route.handler.parent = route

            if not is_class_and_subclass(route.handler, APIView) and not isinstance(
                route.handler, APIView
            ):
                route.handler.create_signature_model()

        if isinstance(route, WebSocketGateway):
            route.handler.create_signature_model(is_websocket=True)

    def validate_root_route_parent(
        self,
        value: Union["Router", "Include", "ASGIApp", "Gateway", "WebSocketGateway"],
        override: bool = False,
    ):
        """
        Handles everything parent from the root. When in the root, the parent must be setup.
        Appends the route path if exists.
        """
        # Getting the value of the router for the path
        value.path = clean_path(self.path + getattr(value, "path", "/"))
        if isinstance(value, (Include, Gateway, WebSocketGateway)):
            if not value.parent and not override:
                value.parent = self

        if isinstance(value, (Gateway, WebSocketGateway)):
            if not is_class_and_subclass(value.handler, APIView) and not isinstance(
                value.handler, APIView
            ):
                if not value.handler.parent:
                    value.handler.parent = value
            else:
                if not value.handler.parent:
                    value(parent=self)

                route_handlers = value.handler.get_route_handlers()
                for route_handler in route_handlers:
                    gateway = (
                        Gateway
                        if not isinstance(route_handler, WebSocketHandler)
                        else WebSocketGateway
                    )
                    gate = gateway(
                        path=value.path,
                        handler=route_handler,
                        name=route_handler.path,
                        middleware=value.middleware,
                        permissions=value.permissions,
                        exception_handlers=value.exception_handlers,
                    )
                    self.routes.append(gate)
                self.routes.pop(self.routes.index(value))


class Router(StarletteRouter, Parent):
    __slots__ = (
        "redirect_slashes",
        "default",
        "name",
        "dependencies",
        "exception_handlers",
        "interceptors",
        "permissions",
        "middleware",
        "response_class",
        "response_cookies",
        "response_headers",
        "parent",
        "tags",
        "deprecated",
        "security",
    )

    def __init__(
        self,
        path: Optional[str] = None,
        app: Optional["Esmerald"] = None,
        parent: Optional["Router"] = None,
        on_shutdown: Optional[List["LifeSpanHandler"]] = None,
        on_startup: Optional[List["LifeSpanHandler"]] = None,
        redirect_slashes: bool = True,
        default: Optional["ASGIApp"] = None,
        routes: Optional[List[Union["APIGateHandler", "Include"]]] = None,
        name: Optional[str] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional["ExceptionHandlers"] = None,
        interceptors: Optional[List["Interceptor"]] = None,
        permissions: Optional[List["Permission"]] = None,
        middleware: Optional[List["Middleware"]] = None,
        response_class: Optional["ResponseType"] = None,
        response_cookies: Optional["ResponseCookies"] = None,
        response_headers: Optional["ResponseHeaders"] = None,
        lifespan: Optional[Callable[["Esmerald"], AsyncContextManager]] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        deprecated: Optional[bool] = None,
        security: Optional[List["SecurityRequirement"]] = None,
    ):
        self.app = app
        if not path:
            path = "/"
        else:
            assert path.startswith("/"), "A path prefix must start with '/'"
            assert not path.endswith(
                "/"
            ), "A path must not end with '/', as the routes will start with '/'"

        for route in routes or []:
            if not isinstance(
                route,
                (
                    Include,
                    Gateway,
                    WebSocketGateway,
                    StarletteBaseRoute,
                    Mount,
                    Host,
                    Router,
                ),
            ):
                raise ImproperlyConfigured(f"The route {route} must be of type Gateway or Include")
        routes = routes or []
        super().__init__(
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            redirect_slashes=redirect_slashes,
            routes=routes,
            default=default,
            lifespan=lifespan,
        )
        self.path = path
        self.parent: Optional["Router"] = parent or self.app
        self.dependencies = dependencies or {}
        self.exception_handlers = exception_handlers or {}
        self.interceptors = interceptors or []
        self.permissions = permissions or []
        self.middleware = middleware or []
        self.routes = routes
        self.tags = tags or []
        self.name = name
        self.response_class = response_class
        self.response_cookies = response_cookies or []
        self.response_headers = response_headers or {}
        self.deprecated = deprecated
        self.security = security or []

        self.routing = copy(self.routes)
        for route in self.routing or []:
            self.validate_root_route_parent(route)

        for route in self.routes or []:
            self.create_signature_models(route)

        self.activate()

    def reorder_routes(self):
        return sorted(
            self.routes,
            key=lambda router: router.path != "" and router.path != "/",
            reverse=True,
        )

    def activate(self):
        self.routes = self.reorder_routes()

    def add_apiview(self, value: "APIView"):
        routes = []
        if not value.handler.parent:
            value(parent=self)

        route_handlers = value.handler.get_route_handlers()
        for route_handler in route_handlers:
            gateway = (
                Gateway if not isinstance(route_handler, WebSocketHandler) else WebSocketGateway
            )
            gate = gateway(
                path=value.path,
                handler=route_handler,
                name=route_handler.path,
                middleware=value.middleware,
                permissions=value.permissions,
                exception_handlers=value.exception_handlers,
            )
            self.routes.append(gate)
            routes.append(gate)

        for route in routes or []:
            self.create_signature_models(route)
        self.activate()

    def add_route(
        self,
        path: str,
        handler: "HTTPHandler",
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional["ExceptionHandlers"] = None,
        interceptors: Optional[List["Interceptor"]] = None,
        permissions: Optional[List["Permission"]] = None,
        middleware: Optional[List["Middleware"]] = None,
        name: Optional[str] = None,
        include_in_schema: bool = True,
        deprecated: Optional[bool] = None,
    ) -> None:
        """
        Adds a new route on a router level making it dynamic based on the handler given.
        If methods is passed, then a new handler will be created for each method passed.
        """
        if not isinstance(handler, HTTPHandler):
            raise ImproperlyConfigured(
                f"handler must be a instance of HTTPHandler and not {handler.__class__.__name__}. "
                "Example: get(), put(), post(), delete(), patch(), route()."
            )
        gateway = Gateway(
            path=self.path + path,
            handler=handler,
            name=name,
            include_in_schema=include_in_schema,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            interceptors=interceptors,
            permissions=permissions,
            middleware=middleware,
            deprecated=deprecated,
        )
        self.validate_root_route_parent(gateway)
        self.create_signature_models(gateway)
        self.routes.append(gateway)

    def add_websocket_route(
        self,
        path: str,
        handler: "WebSocketHandler",
        name: Optional[str] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional["ExceptionHandlers"] = None,
        interceptors: Optional[List["Interceptor"]] = None,
        permissions: Optional[List["Permission"]] = None,
        middleware: Optional[List["Middleware"]] = None,
    ) -> None:
        if not isinstance(handler, WebSocketHandler):
            raise ImproperlyConfigured(
                f"handler must be a instance of WebSocketHandler and not {handler.__class__.__name__}. "
                "Example: websocket()."
            )
        websocket_gateway = WebSocketGateway(
            path=path,
            handler=handler,
            name=name,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            interceptors=interceptors,
            permissions=permissions,
            middleware=middleware,
        )
        self.validate_root_route_parent(websocket_gateway)
        self.create_signature_models(websocket_gateway)
        self.routes.append(websocket_gateway)

    async def not_found(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """Esmerald version of a not found handler when a resource is
        called and cannot be dealt with properly.

        Esmerald overrides the original Starlette not_found to return always
        JSONReponse.

        For plain ASGI apps, just returns the response.
        """
        if scope["type"] == "websocket":
            websocket_close = WebSocketClose()
            await websocket_close(scope, receive, send)
            return

        if "app" in scope:
            raise NotFound(status_code=status.HTTP_404_NOT_FOUND)
        else:
            response = JSONResponse({"detail": "Not Found"}, status_code=status.HTTP_404_NOT_FOUND)
        await response(scope, receive, send)

    def url_path_for(self, name: str, **path_params: Any) -> URLPath:
        for route in self.routes or []:
            try:
                return route.url_path_for(name, **path_params)
            except NoMatchFound:
                pass

            if isinstance(route, (Gateway, WebSocketGateway)):
                try:
                    return route.handler.url_path_for(name, **path_params)
                except NoMatchFound:
                    pass

        raise NoMatchFound(name, path_params)


class HTTPHandler(BaseHandlerMixin, StarletteRoute):
    __slots__ = (
        "path",
        "_permissions",
        "_dependencies",
        "_response_handler",
        "methods",
        "status_code",
        "content_encoding",
        "content_media_type",
        "summary",
        "description",
        "include_in_schema",
        "dependencies",
        "exception_handlers",
        "permissions",
        "middleware",
        "response_class",
        "response_cookies",
        "response_headers",
        "parent",
        "tags",
        "deprecated",
        "security",
        "operation_id",
        "raise_exceptions",
    )

    def __init__(
        self,
        path: Optional[str] = None,
        endpoint: Callable[..., "CoroutineHandler"] = None,
        *,
        methods: Optional[List["HTTPMethod"]] = None,
        status_code: Optional[int] = None,
        content_encoding: Optional[str] = None,
        content_media_type: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        include_in_schema: bool = True,
        background: Optional["BackgroundTaskType"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional["ExceptionHandlers"] = None,
        permissions: Optional[List["Permission"]] = None,
        middleware: Optional[List["Middleware"]] = None,
        media_type: Union[MediaType, str] = MediaType.JSON,
        response_class: Optional["ResponseType"] = None,
        response_cookies: Optional["ResponseCookies"] = None,
        response_headers: Optional["ResponseHeaders"] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        deprecated: Optional[bool] = None,
        response_description: Optional[str] = "Successful Response",
        responses: Optional[Dict[int, ResponseSpecification]] = None,
        security: Optional[List["SecurityRequirement"]] = None,
        operation_id: Optional[str] = None,
        raise_exceptions: Optional[List[Type["HTTPException"]]] = None,
    ) -> None:
        if not path:
            path = "/"
        super().__init__(path=path, endpoint=endpoint, include_in_schema=include_in_schema)
        """
        Handles the "handler" or "apiview" of the platform. A handler can be any get, put, patch, post, delete or route.
        """
        self._permissions: Union[List["Permission"], "VoidType"] = Void
        self._dependencies: Union[Dict[str, Inject], "VoidType"] = Void
        self._response_handler: Union[
            "Callable[[Any], Awaitable[StarletteResponse]]", VoidType
        ] = Void

        self.parent: "ParentType" = None
        self.path = path
        self.endpoint = endpoint
        self.summary = summary
        self.tags = tags or []
        self.include_in_schema = include_in_schema
        self.deprecated = deprecated

        if not self.deprecated:
            self.deprecated = False

        self.security = security or []
        self.operation_id = operation_id

        if not methods:
            methods = [HttpMethod.GET.value]

        for method in methods or []:
            if not isinstance(method, str):
                raise ImproperlyConfigured(f"`{method}` in `methods` must be a string.")

        self.methods: Set[str] = {HttpMethod[method].value for method in methods}
        if HttpMethod.GET in methods:
            self.methods.add(HttpMethod.HEAD.value)

        if isinstance(status_code, IntEnum):
            status_code = int(status_code)
        self.status_code = status_code
        self.status_code = self.handle_status_code()

        self.exception_handlers = exception_handlers or {}
        self.dependencies = dependencies or {}
        self.description = description or inspect.cleandoc(self.endpoint.__doc__ or "")
        self.permissions = list(permissions) if permissions else []
        self.middleware = list(middleware) if middleware else []
        self.description = self.description.split("\f")[0]
        self.media_type = media_type
        self.response_class = response_class
        self.response_cookies = response_cookies
        self.response_headers = response_headers
        self.background = background
        self.signature_model: Optional[Type["SignatureModel"]] = None
        self.transformer: Optional["TransformerModel"] = None
        self.response_description = response_description
        self.responses = responses or {}
        self.content_encoding = content_encoding
        self.content_media_type = content_media_type
        self.raise_exceptions = raise_exceptions

        self.fn: Optional["AnyCallable"] = None
        self.app: Optional["ASGIApp"] = None
        self.route_map: Dict["HTTPMethod" : Tuple["HTTPHandler", "TransformerModel"]] = {}
        self.path_regex, self.path_format, self.param_convertors = compile_path(path)

    def handle_status_code(self):
        """
        Returns the appropriate status code.
        """
        if HttpMethod.GET in self.methods and not self.status_code:
            return status.HTTP_200_OK
        elif HttpMethod.POST in self.methods and not self.status_code:
            return status.HTTP_201_CREATED
        elif HttpMethod.DELETE in self.methods and not self.status_code:
            return status.HTTP_204_NO_CONTENT
        elif HttpMethod.PUT in self.methods and not self.status_code:
            return status.HTTP_200_OK
        elif HttpMethod.PATCH in self.methods and not self.status_code:
            return status.HTTP_200_OK
        elif HttpMethod.HEAD in self.methods and not self.status_code:
            return status.HTTP_200_OK
        return self.status_code

    @property
    def http_methods(self) -> List[str]:
        """
        Converts the methods set into a list of methods.
        """
        return [method for method in self.methods]

    async def allowed_methods(
        self, scope: "Scope", receive: "Receive", send: "Send", methods: List[str]
    ) -> None:
        """
        Validates if the scope method is available within the handler and raises
        a MethodNotAllowed if otherwise.
        """
        for method in methods:
            if method not in self.http_methods:
                raise MethodNotAllowed(detail=f"Method {method.upper()} not allowed.")

    @property
    def allow_header(self):
        """
        Default allow header to be injected in the Response and Starlette Response type
        handlers.
        """
        return {"allow": str(self.methods)}

    @property
    def permission_names(self):
        """
        List of permissions for the route. This is used for OpenAPI Spec purposes only.
        """
        return [permission.__name__ for permission in self.permissions]

    def get_response_class(self) -> Type["Response"]:
        """
        Returns the closest custom Response class in the parent graph or the
        default Response class.
        """
        response_class = Response
        for layer in self.parent_levels:
            if layer.response_class is not None:
                response_class = layer.response_class
        return response_class

    def get_response_headers(self) -> "ResponseHeaders":
        """
        Returns all header parameters in the scope of the handler function.
        """
        resolved_response_headers = {}
        for layer in self.parent_levels:
            resolved_response_headers.update(layer.response_headers or {})
        return resolved_response_headers

    def get_response_cookies(self) -> "ResponseCookies":
        """Returns a list of Cookie instances. Filters the list to ensure each
        cookie key is unique.
        """

        cookies: "ResponseCookies" = []
        for layer in self.parent_levels:
            cookies.extend(layer.response_cookies or [])
        filtered_cookies: "ResponseCookies" = []
        for cookie in reversed(cookies):
            if not any(cookie.key == c.key for c in filtered_cookies):
                filtered_cookies.append(cookie)
        return filtered_cookies

    async def handle(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        ASGIapp that authorizes the connection and then awaits the handler function.
        """
        methods = [scope["method"]]
        await self.allowed_methods(scope, receive, send, methods)

        request = Request(scope=scope, receive=receive, send=send)
        route_handler, parameter_model = self.route_map[scope["method"]]

        if self.get_permissions():
            connection = HTTPConnection(scope=scope, receive=receive)
            await self.allow_connection(connection)

        response = await self.get_response_for_request(
            scope=scope,
            request=request,
            route=route_handler,
            parameter_model=parameter_model,
        )
        await response(scope, receive, send)

    def __call__(self, fn: "AnyCallable") -> "ASGIApp":
        self.fn = fn
        self.endpoint = fn
        self.validate_handler()
        return self

    def check_handler_function(self) -> None:
        """Validates the route handler function once it's set by inspecting its
        return annotations.

        `enable_sync_handlers` from the application settings enables/disables the possibility
        of having `sync` handlers.

        Example:
            enable_sync_handlers = True

            @get(path'/')
            def index(request: Request) -> Response:
                return Response("Welcome Home")

            enable_sync_handlers = False

            @get(path'/')
            async def index(request: Request) -> Response:
                return Response("Welcome Home")

        Raises:
            ImproperlyConfigured if enable_sync_handlers is False and the function is sync.
        """
        if not self.fn:
            raise ImproperlyConfigured(
                "Cannot call check_handler_function without first setting self.fn"
            )

        if not settings.enable_sync_handlers:
            fn = cast("AnyCallable", self.fn)
            if not is_async_callable(fn):
                raise ImproperlyConfigured(
                    "Functions decorated with 'route, websocket, get, patch, put, post and delete' must be async functions"
                )

    def validate_annotations(self):
        """
        Validate annotations of the handlers.
        """
        return_annotation = self.signature.return_annotation

        if return_annotation is Signature.empty:
            raise ImproperlyConfigured(
                "A return value of a route handler function should be type annotated."
                "If your function doesn't return a value or returns None, annotate it as returning 'NoReturn' or 'None' respectively."
            )
        if (
            self.status_code < status.HTTP_200_OK
            or self.status_code in {status.HTTP_204_NO_CONTENT, status.HTTP_304_NOT_MODIFIED}
        ) and return_annotation not in [NoReturn, None]:
            raise ImproperlyConfigured(
                "A status code 204, 304 or in the range below 200 does not support a response body."
                " If the function should return a value, change the route handler status code to an appropriate value.",
            )
        if (
            is_class_and_subclass(return_annotation, Redirect)
            and self.status_code not in REDIRECT_STATUS_CODES
        ):
            raise ValidationErrorException(
                f"Redirect responses should have one of "
                f"the following status codes: {', '.join([str(s) for s in REDIRECT_STATUS_CODES])}"
            )
        if is_class_and_subclass(return_annotation, File) and self.media_type in [
            MediaType.JSON,
            MediaType.HTML,
        ]:
            self.media_type = MediaType.TEXT

    def validate_reserved_kwargs(self):
        """
        Validates if special words are in the signature.
        """
        if DATA in self.signature.parameters and "GET" in self.methods:
            raise ImproperlyConfigured("'data' argument is unsupported for 'GET' request handlers")

        if SOCKET in self.signature.parameters:
            raise ImproperlyConfigured("The 'socket' argument is not supported with http handlers")

    def validate_handler(self):
        self.check_handler_function()
        self.validate_annotations()
        self.validate_reserved_kwargs()

    async def to_response(self, app: "Esmerald", data: Any) -> StarletteResponse:
        response_handler = self.get_response_handler()
        return await response_handler(app=app, data=data)


class WebSocketHandler(BaseHandlerMixin, StarletteWebSocketRoute):
    """
    Websocket handler object representation.
    """

    __slots__ = (
        "path",
        "dependencies",
        "exception_handlers",
        "permissions",
        "middleware",
        "parent",
    )

    def __init__(
        self,
        path: Union[Optional[str], Optional[List[str]]] = None,
        *,
        endpoint: Callable[..., "CoroutineHandler"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional[Dict[Union[int, Type[Exception]], "ExceptionHandler"]] = None,
        permissions: Optional[List["Permission"]] = None,
        middleware: Optional[List["Middleware"]] = None,
    ):
        if not path:
            path = "/"
        super().__init__(path=path, endpoint=endpoint)
        self._permissions: Union[List["Permission"], "VoidType"] = Void
        self._dependencies: Union[Dict[str, Inject], "VoidType"] = Void
        self._response_handler: Union[
            "Callable[[Any], Awaitable[StarletteResponse]]", VoidType
        ] = Void

        self.parent: "ParentType" = None
        self.dependencies = dependencies
        self.exception_handlers = exception_handlers
        self.permissions = permissions
        self.middleware = middleware
        self.signature_model: Optional[Type["SignatureModel"]] = None
        self.websocket_parameter_model: Optional["TransformerModel"] = None

    def __call__(self, fn: "AnyCallable") -> "ASGIApp":
        self.fn = fn
        self.endpoint = fn
        self.validate_websocket_handler_function()
        return self

    def validate_reserved_words(self, signature: "Signature"):
        """
        Validates if special words are in the signature.
        """
        if signature.return_annotation is not None:
            raise ImproperlyConfigured("Websocket functions should return 'None'.")

        unsupported_kwargs = [REQUEST, DATA]
        for kwarg in unsupported_kwargs:
            if kwarg in signature.parameters:
                raise ImproperlyConfigured(
                    f"The '{kwarg}'is not supported with websocket handlers."
                )

    def validate_websocket_handler_function(self) -> None:
        """
        Validates the route handler function once it is set by inspecting its
        return annotations.
        """
        if not self.fn:
            raise ImproperlyConfigured(
                "Cannot call check_handler_function without first setting self.fn"
            )
        signature = Signature.from_callable(self.fn)
        self.validate_reserved_words(signature=signature)

        if SOCKET not in signature.parameters:
            raise ImproperlyConfigured("Websocket handlers must set a 'socket' argument.")
        if not is_async_callable(self.fn):
            raise ImproperlyConfigured(
                "Functions decorated with 'asgi, get, patch, put, post and delete' must be async functions."
            )

    async def handle(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """The handle of a websocket"""
        websocket = WebSocket[Any, Any](scope=scope, receive=receive, send=send)
        if self.get_permissions():
            await self.allow_connection(connection=websocket)

        kwargs = await self.get_kwargs(websocket=websocket)

        fn = cast("AsyncAnyCallable", self.fn)
        if isinstance(self.parent, APIView):
            await fn(self.parent, **kwargs)
        else:
            await fn(**kwargs)

    async def get_kwargs(self, websocket: WebSocket[Any, Any]) -> Dict[str, Any]:
        """Resolves the required kwargs from the request data.

        Args:
            websocket: WebSocket instance

        Returns:
            Dictionary of parsed kwargs
        """
        assert self.websocket_parameter_model, "handler parameter model not defined."

        signature_model = get_signature(self)
        kwargs = self.websocket_parameter_model.to_kwargs(connection=websocket)
        for dependency in self.websocket_parameter_model.dependencies:
            kwargs[dependency.key] = await self.websocket_parameter_model.get_dependencies(
                dependency=dependency, connection=websocket, **kwargs
            )
        return signature_model.parse_values_for_connection(connection=websocket, **kwargs)


class Include(Mount):
    """
    Esmerald version of the classic mount from Starlette but extends it further
    to contemplate the includes.

    Handle with specific includes from the URLs.

    Include manages routes as a list or as a namespace but not both or a
    ImproperlyConfigured is raised.

    If the app is an Esmerald or ChildEsmerald, then the parent is set to be the include
    itself.
    """

    __slots__ = (
        "path",
        "app",
        "namespace",
        "pattern",
        "name",
        "dependencies",
        "exception_handlers",
        "interceptors",
        "permissions",
        "middleware",
        "deprecated",
        "security",
    )

    def __init__(
        self,
        path: Optional[str] = None,
        app: Optional["ASGIApp"] = None,
        name: Optional[str] = None,
        routes: Optional[List[StarletteBaseRoute]] = None,
        namespace: Optional[str] = None,
        pattern: Optional[str] = None,
        parent: Optional["Router"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional[Dict[Union[int, Type[Exception]], "ExceptionHandler"]] = None,
        interceptors: Optional[List["Interceptor"]] = None,
        permissions: Optional[List["Permission"]] = None,
        middleware: Optional[List["Middleware"]] = None,
        include_in_schema: Optional[bool] = True,
        deprecated: Optional[bool] = None,
        security: Optional[List["SecurityRequirement"]] = None,
    ) -> None:
        self.path = path
        if not path:
            self.path = "/"

        if namespace and routes:
            raise ImproperlyConfigured("It can only be namespace or routes, not both.")

        if namespace and not isinstance(namespace, str):
            raise ImproperlyConfigured("Namespace must be a string. Example: 'myapp.routes'.")

        if pattern and not isinstance(pattern, str):
            raise ImproperlyConfigured("Pattern must be a string. Example: 'route_patterns'.")

        if pattern and routes:
            raise ImproperlyConfigured("Pattern must be used only with namespace.")

        if namespace:
            routes: List["APIGateHandler"] = include(namespace, pattern)

        self.app = app
        self.name = name
        self.namespace = namespace
        self.pattern = pattern
        self.include_in_schema = include_in_schema
        self.dependencies = dependencies or {}
        self.interceptors = interceptors or []
        self.permissions = permissions or []
        self.middleware = middleware or []
        self.exception_handlers = exception_handlers or {}
        self.deprecated = deprecated
        self.response_class = None
        self.response_cookies = None
        self.response_headers = None
        self.parent = parent
        self.security = security or []

        if routes:
            routes = self.resolve_route_path_handler(routes)

        # Build the middleware from the routes
        routes_middleware = []
        for route in routes or []:
            routes_middleware = self.build_routes_middleware(route)

        # Add the middleware to the include
        self.middleware += routes_middleware
        include_middleware = []

        if self.middleware:
            for middleware in self.middleware:
                if isinstance(middleware, StarletteMiddleware):
                    include_middleware.append(middleware)
                else:
                    include_middleware.append(StarletteMiddleware(middleware))

        super().__init__(
            self.path, app=self.app, routes=routes, name=name, middleware=include_middleware
        )

        self.path_regex, self.path_format, self.param_convertors = compile_path(
            self.path + "/{path:path}"
        )

    def build_routes_middleware(
        self, route: "RouteParent", middlewares: Optional[List["Middleware"]] = None
    ):
        """
        Builds the middleware stack from the top to the bottom of the routes.
        """
        from esmerald import ChildEsmerald, Esmerald

        if not middlewares:
            middlewares = []

        if isinstance(route, Include):
            app = getattr(route, "app", None)
            if app and isinstance(app, (Esmerald, ChildEsmerald)):
                return middlewares

        if isinstance(route, (Gateway, WebSocketGateway)):
            middlewares.extend(route.middleware)
            if route.handler.middleware:
                middlewares.extend(route.handler.middleware)

        return middlewares

    def resolve_route_path_handler(self, routes: List[StarletteBaseRoute]):
        """
        Make sure the paths are properly configured from the handler endpoint.
        The handler can be a Starlette function, an APIView or a HTTPHandler.

        Mount() has a limitation of not allowing nested Mount().

        Example:

            route_patterns = [
                Mount(path='/my path', routes=[
                    Mount(path='/another mount')
                ])
            ]


        Include() extends the Mount and adds extras on the top. Allowing nested
        Include() also allows importing in different ways. Via qualified namespace
        or via list() but not both.

        Example:

            1. Root of the application, example, `urls.py`.

            route_patterns = [
                Include(path='/api/v1/', namespace='myapp.v1.urls'),
                Gateway(path='/example', handler=example_endpoint, name='example')
                ...
            ]

            2. Inside `myapp.v1.urls`

            from mysecondapp.urls import route_patterns

            route_patterns = [
                Include(path='/api/v2/', routes=route_patterns),
                Gateway(path='/another-example', handler=another_endpoint, name='example')
                ...
            ]

            3. Inside `mysecondapp.v1.urls`:

            route_patterns = [
                Gateway(path='/last-example', handler=another_endpoint, name='example')
                ...
            ]

        """
        routing = []

        for route in routes:
            if not isinstance(route, (Include, Gateway, WebSocketGateway, Mount)):
                raise ImproperlyConfigured("The route must be of type Gateway or Include")
            route.parent = self
            if isinstance(route, Include):
                routing.append(route)
            else:
                if isinstance(route.handler, (HTTPHandler, WebSocketHandler)):
                    route.handler.parent = route
                    routing.append(route)
                    continue

                if is_class_and_subclass(route.handler, APIView) or isinstance(
                    route.handler, APIView
                ):
                    if not route.handler.parent:
                        route.handler = route.handler(parent=self)
                    route_handlers = route.handler.get_route_handlers()

                    for route_handler in route_handlers:
                        gateway = (
                            Gateway
                            if not isinstance(route_handler, WebSocketHandler)
                            else WebSocketGateway
                        )
                        gate = gateway(
                            path=route.path,
                            handler=route_handler,
                            name=route_handler.path,
                            middleware=route.middleware,
                            interceptors=self.interceptors,
                            permissions=route.permissions,
                            exception_handlers=route.exception_handlers,
                        )

                        routing.append(gate)
        return routing



"""
Signature is widely used by Pydantic and comes from the inpect library.
A lot of great work was done using the Signature and Esmerald is no exception.
"""

from inspect import Parameter
from inspect import Signature as InspectSignature
from typing import Any, Generator, Set, TYPE_CHECKING, Type

from esmerald.exceptions import ImproperlyConfigured
from esmerald.parsers import BaseModelExtra
from esmerald.transformers.constants import CLASS_SPECIAL_WORDS, VALIDATION_NAMES
from esmerald.transformers.datastructures import EsmeraldSignature, Parameter
from esmerald.transformers.helpers import is_pydantic_constrained_field
from esmerald.transformers.utils import get_field_definition_from_param
from esmerald.utils.dependency import (
    is_dependency_field,
    should_skip_dependency_validation,
)
from pydantic import create_model
from pydantic.fields import Undefined

if TYPE_CHECKING:
    from pydantic.typing import AnyCallable, DictAny


class SignatureFactory(BaseModelExtra):
    class Config(BaseModelExtra.Config):
        arbitrary_types_allowed = True

    def __init__(self, fn: "AnyCallable", dependency_names: Set[str], **kwargs: "DictAny") -> None:
        super().__init__(**kwargs)
        if not fn or fn is None:
            raise ImproperlyConfigured("Parameter 'fn' to SignatureFactory cannot be `None`.")
        self.fn = fn
        self.signature = InspectSignature.from_callable(self.fn)
        self.fn_name = fn.__name__ if hasattr(fn, "__name__") else "anonymous"
        self.defaults = {}
        self.dependency_names = dependency_names
        self.field_definitions = {}

    def validate_missing_dependency(self, param: Parameter) -> None:
        if param.optional:
            return
        if not is_dependency_field(param.default):
            return
        field = param.default
        if field.default is not Undefined:
            return
        if param.name not in self.dependency_names:
            raise ImproperlyConfigured(
                f"Explicit dependency '{param.name}' for '{self.fn_name}' has no default value, "
                f"or provided dependency."
            )

    def get_dependency_names(self, param: Parameter) -> None:
        if is_dependency_field(param.default):
            self.dependency_names.add(param.name)

    def set_default_field(self, param: Parameter) -> None:
        if param.default_defined:
            self.defaults[param.name] = param.default

    @property
    def parameters(self) -> Generator[Parameter, None, None]:
        for name, param in self.signature.parameters.items():
            if name in CLASS_SPECIAL_WORDS:
                continue
            yield Parameter(self.fn_name, name, param)

    def skip_parameter_validation(self, param: Parameter) -> bool:
        return param.name in VALIDATION_NAMES or should_skip_dependency_validation(param.default)

    def create_signature(self) -> Type[EsmeraldSignature]:
        try:
            for param in self.parameters:
                self.validate_missing_dependency(param)
                self.get_dependency_names(param)
                self.set_default_field(param)

                if self.skip_parameter_validation(param):
                    self.field_definitions[param.name] = (Any, ...)
                    continue
                if is_pydantic_constrained_field(param.default):
                    self.field_definitions[param.name] = (param.default, ...)
                    continue
                self.field_definitions[param.name] = get_field_definition_from_param(param)

            model: Type["EsmeraldSignature"] = create_model(
                self.fn_name + "_signature",
                __base__=EsmeraldSignature,
                **self.field_definitions,
            )
            model.return_annotation = self.signature.return_annotation
            model.dependency_names = self.dependency_names
            return model
        except TypeError as e:
            raise ImproperlyConfigured(
                f"Error creating signature for '{self.fn_name}': '{e}'."
            ) from e



from datetime import datetime, timedelta
from typing import List, Union

from pydantic import BaseModel


class JWTConfig(BaseModel):
    signing_key: str
    api_key_header: str = "X_API_TOKEN"
    algorithm: str = "HS256"
    access_token_lifetime: Union[datetime, timedelta, str, float] = timedelta(minutes=5)
    refresh_token_lifetime: Union[datetime, timedelta, str, float] = timedelta(days=1)
    auth_header_types: List[str] = ["Bearer"]
    jti_claim: str = "jti"
    verifying_key: str = ""
    leeway: str = 0
    sliding_token_lifetime: Union[datetime, timedelta, str, float] = timedelta(minutes=5)
    sliding_token_refresh_lifetime: Union[datetime, timedelta, str, float] = timedelta(days=1)
    user_id_field: str = "id"
    user_id_claim: str = "user_id"
    access_token_name: str = "access_token"
    refresh_token_name: str = "refresh_token"



import re
from typing import TYPE_CHECKING, List, Optional, Union

from starlette.routing import Route as StarletteRoute
from starlette.routing import WebSocketRoute as StarletteWebSocketRoute
from starlette.routing import compile_path
from starlette.types import Receive, Scope, Send

from esmerald.routing.base import BaseInterceptorMixin
from esmerald.routing.views import APIView
from esmerald.typing import Void
from esmerald.utils.helpers import clean_string, is_class_and_subclass
from esmerald.utils.url import clean_path

if TYPE_CHECKING:
    from esmerald.interceptors.types import Interceptor
    from esmerald.permissions.types import Permission
    from esmerald.routing.router import HTTPHandler, WebSocketHandler
    from esmerald.routing.views import APIView
    from esmerald.types import Dependencies, ExceptionHandlers, Middleware, ParentType


class Gateway(StarletteRoute, BaseInterceptorMixin):
    __slots__ = (
        "_interceptors",
        "path",
        "handler",
        "name",
        "include_in_schema",
        "parent",
        "dependencies",
        "middleware",
        "exception_handlers",
        "interceptors",
        "permissions",
        "deprecated",
    )

    def __init__(
        self,
        path: Optional[str] = None,
        *,
        handler: Union["HTTPHandler", "APIView"],
        name: Optional[str] = None,
        include_in_schema: bool = True,
        parent: Optional["ParentType"] = None,
        dependencies: Optional["Dependencies"] = None,
        middleware: Optional["Middleware"] = None,
        interceptors: Optional["Interceptor"] = None,
        permissions: Optional["Permission"] = None,
        exception_handlers: Optional["ExceptionHandlers"] = None,
        deprecated: Optional[bool] = None,
        is_from_router: bool = False,
    ) -> None:
        if not path:
            path = "/"
        if is_class_and_subclass(handler, APIView):
            handler = handler(parent=self)

        if not is_from_router:
            self.path = clean_path(path + handler.path)
        else:
            self.path = clean_path(path)
        self.methods = getattr(handler, "methods", None)

        if not name:
            if not isinstance(handler, APIView):
                name = clean_string(handler.fn.__name__)
            else:
                name = clean_string(handler.__class__.__name__)

        super().__init__(
            path=self.path,
            endpoint=handler,
            include_in_schema=include_in_schema,
            name=name,
            methods=self.methods,
        )
        """
        A "bridge" to a handler and router mapping functionality.
        Since the default Starlette Route endpoint does not understand the Esmerald handlers,
        the Gateway bridges both functionalities and adds an extra "flair" to be compliant with both class based views and decorated function views.
        """
        self._interceptors: Union[List["Interceptor"], "Void"] = Void

        self.handler = handler
        self.dependencies = dependencies or {}
        self.interceptors = interceptors or []
        self.permissions = permissions or []
        self.middleware = middleware or []
        self.exception_handlers = exception_handlers or {}
        self.response_class = None
        self.response_cookies = None
        self.response_headers = None
        self.deprecated = deprecated
        self.parent = parent
        (
            handler.path_regex,
            handler.path_format,
            handler.param_convertors,
        ) = compile_path(self.path)

        if not is_class_and_subclass(self.handler, APIView) and not isinstance(
            self.handler, APIView
        ):
            self.handler.get_response_handler()

            if not handler.operation_id:
                handler.operation_id = self.generate_operation_id()

    async def handle(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Handles the interception of messages and calls from the API.
        """
        if self.get_interceptors():
            await self.intercept(scope, receive, send)

        await self.handler.handle(scope, receive, send)

    def generate_operation_id(self):
        """
        Generates an unique operation if for the handler
        """
        operation_id = self.name + self.handler.path_format
        operation_id = re.sub(r"\W", "_", operation_id)
        methods = [method for method in self.handler.methods]
        operation_id = f"{operation_id}_{methods[0].lower()}"
        return operation_id


class WebSocketGateway(StarletteWebSocketRoute, BaseInterceptorMixin):
    __slots__ = (
        "_interceptors",
        "path",
        "handler",
        "name",
        "dependencies",
        "middleware",
        "exception_handlers",
        "interceptors",
        "permissions",
        "parent",
    )

    def __init__(
        self,
        path: Optional[str] = None,
        *,
        handler: Union["WebSocketHandler", "APIView"],
        name: Optional[str] = None,
        parent: Optional["ParentType"] = None,
        dependencies: Optional["Dependencies"] = None,
        middleware: Optional[List["Middleware"]] = None,
        exception_handlers: Optional["ExceptionHandlers"] = None,
        interceptors: Optional[List["Interceptor"]] = None,
        permissions: Optional[List["Permission"]] = None,
    ) -> None:
        if not path:
            path = "/"
        if is_class_and_subclass(handler, APIView):
            handler = handler(parent=self)
        self.path = clean_path(path + handler.path)

        if not name:
            if not isinstance(handler, APIView):
                name = clean_string(handler.fn.__name__)
            else:
                name = clean_string(handler.__class__.__name__)

        super().__init__(
            path=self.path,
            endpoint=handler,
            name=name,
        )
        """
        A "bridge" to a handler and router mapping functionality.
        Since the default Starlette Route endpoint does not understand the Esmerald handlers,
        the Gateway bridges both functionalities and adds an extra "flair" to be compliant with both class based views and decorated function views.
        """
        self._interceptors: Union[List["Interceptor"], "Void"] = Void
        self.handler = handler
        self.dependencies = dependencies or {}
        self.interceptors = interceptors or []
        self.permissions = permissions or []
        self.middleware = middleware or []
        self.exception_handlers = exception_handlers or {}
        self.parent = parent
        (
            handler.path_regex,
            handler.path_format,
            handler.param_convertors,
        ) = compile_path(self.path)

    async def handle(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Handles the interception of messages and calls from the API.
        """
        if self.get_interceptors():
            await self.intercept(scope, receive, send)

        await self.handler.handle(scope, receive, send)



import os
import sys
from collections import defaultdict
from difflib import get_close_matches

import esmerald
from esmerald.core.management.base import BaseDirective, DirectiveError, DirectiveParser
from esmerald.core.management.utils import get_directives, load_directive_class


class ManagementUtility:
    location = __path__[0]

    def __init__(self, argv=None):
        self.argv = argv or sys.argv[:]
        self.prog_name = os.path.basename(self.argv[0])

    def main_help_text(self, directives_only=False):
        if directives_only:
            usage = sorted(get_directives(self.location))
        else:
            usage = [
                "",
                "Type '%s help <subdirective>' for help on a specific subdirective."
                % self.prog_name,
                "",
                "Available subdirectives:",
            ]
            directives_dict = defaultdict(lambda: [])
            for name, app in get_directives().items():
                if app == "esmerald.core":
                    app = "esmerald"
                else:
                    app = app.rpartition(".")[-1]
                directives_dict[app].append(name)

        return "\n".join(usage)

    def fetch_directive(self, subdirective):
        directives = get_directives(self.location)
        try:
            app_name = directives[subdirective]
        except KeyError:
            possible_matches = get_close_matches(subdirective, directives)
            sys.stderr.write("Unknown command: %r" % subdirective)
            if possible_matches:
                sys.stderr.write(". Did you mean %s?" % possible_matches[0])
            sys.stderr.write("\nType '%s help' for usage.\n" % self.prog_name)
            sys.exit(1)
        if isinstance(app_name, BaseDirective):
            # If the command is already loaded, use it directly.
            klass = app_name
        else:
            klass = load_directive_class(app_name, subdirective)
        return klass

    def run(self):
        try:
            subdirective = self.argv[1]
        except IndexError:
            subdirective = "help"

        parser = DirectiveParser(
            prog=self.prog_name,
            usage="%(prog)s subdirective [options] [args]",
            add_help=False,
            allow_abbrev=False,
        )
        parser.add_argument("args", nargs="*")
        try:
            options, args = parser.parse_known_args(self.argv[2:])
        except DirectiveError:
            pass

        if subdirective == "help":
            if "--directives" in args:
                sys.stdout.write(self.main_help_text(directives_only=True) + "\n")
            elif not options.args:
                sys.stdout.write(self.main_help_text() + "\n")
            else:
                self.fetch_directive(options.args[0]).print_help(self.prog_name, options.args[0])
        elif subdirective == "version" or self.argv[1:] == ["--version"]:
            sys.stdout.write(esmerald.__version__ + "\n")
        elif self.argv[1:] in (["--help"], ["-h"]):
            sys.stdout.write(self.main_help_text() + "\n")
        else:
            self.fetch_directive(subdirective).run_from_argv(self.argv)


def run_from_command_line(argv=None):
    """Run a ManagementUtility."""
    utility = ManagementUtility(argv)
    utility.run()



from esmerald.conf import settings as esmerald_settings
from esmerald.protocols.middleware import MiddlewareProtocol
from esmerald.types import ASGIApp, Receive, Scope, Send


class RequestSettingsMiddleware(MiddlewareProtocol):
    def __init__(self, app: "ASGIApp"):
        """Settings Middleware class.

        Args:
            app: The 'next' ASGI app to call.
            settings: An instance of a EsmeraldAPISettings or a subclass.
        """
        super().__init__(app)
        self.app = app

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        app = scope["app"]
        scope["global_settings"] = esmerald_settings
        scope["app_settings"] = app.settings_config if app.settings_config else None

        await self.app(scope, receive, send)



from tortoise import fields, models

from .base_user import (
    AbstractUser,
    AutoIncrementBigIntMixin,
    AutoIncrementIntMixin,
    User,
)

__all__ = [
    "User",
    "AbstractUser",
    "AutoIncrementBigIntMixin",
    "AutoIncrementIntMixin",
    "fields",
    "models",
]



from typing import Union

from pydantic import BaseConfig, BaseModel, constr, validator
from typing_extensions import Literal

from esmerald.datastructures import Secret

SECONDS_IN_A_DAY = 60 * 60 * 24


class SessionConfig(BaseModel):
    class Config(BaseConfig):
        arbitrary_types_allowed = True

    secret_key: Union[str, Secret]
    path: str = "/"
    session_cookie: constr(min_length=1, max_length=256) = "session"  # type: ignore[valid-type]
    max_age: int = SECONDS_IN_A_DAY * 180  # type: ignore[valid-type]
    https_only: bool = False
    same_site: Literal["lax", "strict", "none"] = "lax"

    @validator("secret_key", always=True)
    def validate_secret(cls, value: Secret) -> Secret:  # pylint: disable=no-self-argument
        if len(value) not in [16, 24, 32]:
            raise ValueError("secret length must be 16 (128 bit), 24 (192 bit) or 32 (256 bit)")
        return value



from typing import Any, Callable, List, Optional, TypeVar

from starlette.background import BackgroundTask as StarletteBackgroundTask  # noqa
from starlette.background import BackgroundTasks as StarletteBackgroundTasks  # noqa
from starlette.responses import Response as StarletteResponse  # noqa
from typing_extensions import ParamSpec

P = ParamSpec("P")
R = TypeVar("R", bound=StarletteResponse)


class BackgroundTask(StarletteBackgroundTask):
    def __init__(self, func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> None:
        super().__init__(func, *args, **kwargs)


class BackgroundTasks(StarletteBackgroundTasks):
    def __init__(self, tasks: Optional[List[BackgroundTask]] = None):
        super().__init__(tasks=tasks)



from typing import Type, Union

from pydantic import (
    ConstrainedBytes,
    ConstrainedDate,
    ConstrainedDecimal,
    ConstrainedFloat,
    ConstrainedFrozenSet,
    ConstrainedInt,
    ConstrainedList,
    ConstrainedSet,
    ConstrainedStr,
)
from typing_extensions import TypeGuard

ConstrainedField = TypeGuard[
    Union[
        Type[ConstrainedBytes],
        Type[ConstrainedDate],
        Type[ConstrainedDecimal],
        Type[ConstrainedFloat],
        Type[ConstrainedFrozenSet],
        Type[ConstrainedInt],
        Type[ConstrainedList],
        Type[ConstrainedSet],
        Type[ConstrainedStr],
    ]
]



from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, cast

from esmerald.enums import HttpMethod
from esmerald.openapi.parameters import create_parameter_for_handler
from esmerald.openapi.request_body import create_request_body
from esmerald.openapi.responses import create_responses
from openapi_schemas_pydantic.v3_1_0.operation import Operation
from openapi_schemas_pydantic.v3_1_0.path_item import PathItem
from starlette.routing import get_name

if TYPE_CHECKING:
    from esmerald.routing.handlers import HTTPHandler
    from esmerald.types import AnyCallable
    from openapi_schemas_pydantic.v3_1_0 import SecurityRequirement
    from pydantic import BaseModel


def get_description_for_handler(
    handler: "HTTPHandler", use_handler_docstrings: bool
) -> Optional[str]:
    description = handler.description
    if description is None and use_handler_docstrings:
        return handler.fn.__doc__
    return description


def extract_level_values(
    handler: "HTTPHandler",
) -> Tuple[Optional[List[str]], Optional[List[Dict[str, List[str]]]]]:
    tags: List[str] = []
    security: List["SecurityRequirement"] = []

    for layer in handler.parent_levels:
        if hasattr(layer, "tags"):
            tags.extend(layer.tags or [])
        if hasattr(layer, "security"):
            security.extend(layer.security or [])
    return list(set(tags)) if tags else None, security or None


def create_path_item(
    route: "HTTPHandler", create_examples: bool, use_handler_docstrings: bool
) -> PathItem:
    path_item = PathItem()

    # remove the HEAD from the docs
    route_map = {k: v for k, v in route.route_map.items() if k != HttpMethod.HEAD}

    for http_method, handler_tuple in route_map.items():
        handler, _ = handler_tuple

        if handler.include_in_schema:
            handler_fields = cast("BaseModel", handler.signature_model).__fields__
            parameters = (
                create_parameter_for_handler(
                    handler=handler,
                    handler_fields=handler_fields,
                    path_parameters=handler.normalised_path_params,
                    create_examples=create_examples,
                )
                or None
            )
            raises_validation_error = bool(
                "data" in handler_fields or path_item.parameters or parameters
            )
            handler_name = get_name(cast("AnyCallable", handler.fn)).replace("_", " ").title()
            request_body = None

            if "data" in handler_fields:
                request_body = create_request_body(
                    field=handler_fields["data"], create_examples=create_examples
                )

            tags, security = extract_level_values(handler=handler)
            operation = Operation(
                operationId=handler.operation_id or handler_name,
                tags=tags,
                summary=handler.summary,
                description=get_description_for_handler(handler, use_handler_docstrings),
                deprecated=handler.deprecated,
                responses=create_responses(
                    handler=handler,
                    raises_validation_error=raises_validation_error,
                    create_examples=create_examples,
                ),
                requestBody=request_body,
                parameters=parameters,
                security=security,
            )
            setattr(path_item, http_method.lower(), operation)
    return path_item



from typing import TYPE_CHECKING, List, Optional  # noqa

from esmerald.exceptions import PermissionDenied  # noqa

if TYPE_CHECKING:
    from esmerald.permissions.types import Permission  # noqa
    from esmerald.requests import Request  # noqa
    from esmerald.types import APIGateHandler  # noqa


def check_permissions(
    request: "Request",
    apiview: "APIGateHandler",
    permissions: List["Permission"],
) -> None:
    """
    Check if the request should be permitted.
    Raises an appropriate exception if the request is not permitted.
    """
    for permission in permissions:
        if not permission().has_permission(request, apiview):
            permission_denied(
                request,
                message=getattr(permission, "message", None),
            )


def continue_or_raise_permission_exception(
    request: "Request",
    apiview: "APIGateHandler",
    permission: "Permission",
):
    if not permission.has_permission(request, apiview):
        permission_denied(
            request,
            message=getattr(permission, "message", None),
        )


def permission_denied(request: "Request", message: Optional[str] = None) -> PermissionDenied:
    """
    If request is not permitted, determine what kind of exception to raise.
    """
    raise PermissionDenied(detail=message, status_code=403)



from typing import TYPE_CHECKING, Callable, Dict

from esmerald.enums import MediaType, OpenAPIMediaType
from esmerald.exceptions import ImproperlyConfigured
from esmerald.requests import Request
from esmerald.responses import Response
from esmerald.routing.handlers import get
from esmerald.routing.views import APIView
from orjson import OPT_INDENT_2, dumps
from starlette.status import HTTP_200_OK, HTTP_404_NOT_FOUND

if TYPE_CHECKING:
    from openapi_schemas_pydantic.v3_1_0.open_api import OpenAPI
    from typing_extensions import Literal

MSG_OPENAPI_NOT_INITIALIZED = "Esmerald has not been created with an OpenAPIConfig"


class OpenAPIView(APIView):
    """
    The view that manages the OpenAPI documentation
    """

    path: str = "/docs"
    style: str = "body { margin: 0; padding: 0 }"
    redoc_version: str = "next"
    swagger_ui_version: str = "4.14.0"
    stoplight_elements_version: str = "7.6.5"
    favicon_url: str = "https://esmerald.dymmond.com/statics/images/favicon.ico"
    redoc_google_fonts: bool = True
    redoc_js_url: str = (
        f"https://cdn.jsdelivr.net/npm/redoc@{redoc_version}/bundles/redoc.standalone.js"
    )
    swagger_css_url: str = (
        f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{swagger_ui_version}/swagger-ui.css"
    )
    swagger_ui_bundle_js_url: str = (
        f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{swagger_ui_version}/swagger-ui-bundle.js"
    )
    swagger_ui_standalone_preset_js_url: str = f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{swagger_ui_version}/swagger-ui-standalone-preset.js"
    stoplight_elements_css_url: str = (
        f"https://unpkg.com/@stoplight/elements@{stoplight_elements_version}/styles.min.css"
    )
    stoplight_elements_js_url: str = (
        f"https://unpkg.com/@stoplight/elements@{stoplight_elements_version}/web-components.min.js"
    )

    _dumped_schema: str = ""
    _dumped_modified_schema: str = ""

    @staticmethod
    def get_schema_from_request(request: Request) -> "OpenAPI":
        if not request.app.openapi_schema:
            raise ImproperlyConfigured(MSG_OPENAPI_NOT_INITIALIZED)
        return request.app.openapi_schema

    def should_serve_endpoint(self, request: Request) -> bool:
        if not request.app.openapi_config:
            raise ImproperlyConfigured(MSG_OPENAPI_NOT_INITIALIZED)

        request_path = set(filter(None, request.url.path.split("/")))
        root_path = request.app.root_path or set(filter(None, self.path.split("/")))

        config = request.app.openapi_config

        if request_path == root_path and config.root_schema_site in config.enabled_endpoints:
            return True

        if request_path & config.enabled_endpoints:
            return True

        return False

    @property
    def favicon(self) -> str:
        return (
            f"<link rel='icon' type='image/x-icon' href='{self.favicon_url}'>"
            if self.favicon_url
            else "<meta/>"
        )

    @property
    def render_methods_map(
        self,
    ) -> Dict["Literal['redoc', 'swagger', 'elements']", Callable[[Request], str]]:
        """
        Returns:
            A mapping of string keys to render methods.
        """
        return {
            "redoc": self.render_redoc,
            "swagger": self.render_swagger_ui,
            "elements": self.render_stoplight_elements,
        }

    @get(
        path="/openapi.yaml",
        media_type=OpenAPIMediaType.OPENAPI_YAML,
        include_in_schema=False,
    )
    def retrieve_schema_yaml(self, request: Request) -> Response:

        if not request.app.openapi_config:  # pragma: no cover
            raise ImproperlyConfigured(MSG_OPENAPI_NOT_INITIALIZED)

        if self.should_serve_endpoint(request):
            return Response(
                content=self.get_schema_from_request(request),
                status_code=HTTP_200_OK,
                media_type=OpenAPIMediaType.OPENAPI_YAML,
            )
        return Response(
            content={},
            status_code=HTTP_404_NOT_FOUND,
            media_type=MediaType.JSON,
        )

    @get(
        path="/openapi.json",
        media_type=OpenAPIMediaType.OPENAPI_JSON,
        include_in_schema=False,
    )
    def retrieve_schema_json(self, request: Request) -> Response:
        if not request.app.openapi_config:  # pragma: no cover
            raise ImproperlyConfigured(MSG_OPENAPI_NOT_INITIALIZED)

        if self.should_serve_endpoint(request):
            return Response(
                content=self.get_schema_from_request(request),
                status_code=HTTP_200_OK,
                media_type=OpenAPIMediaType.OPENAPI_JSON,
            )
        return Response(
            content={},
            status_code=HTTP_404_NOT_FOUND,
            media_type=MediaType.JSON,
        )

    @get(path="/", media_type=MediaType.HTML, include_in_schema=False)
    def root(self, request: Request) -> Response:
        config = request.app.openapi_config
        if not config:  # pragma: no cover
            raise ImproperlyConfigured(MSG_OPENAPI_NOT_INITIALIZED)
        render_method = self.render_methods_map[config.root_schema_site]

        if self.should_serve_endpoint(request):
            return Response(
                content=render_method(request),
                status_code=HTTP_200_OK,
                media_type=MediaType.HTML,
            )
        return Response(
            content=self.render_404_page(),
            status_code=HTTP_404_NOT_FOUND,
            media_type=MediaType.HTML,
        )

    @get(path="/swagger", media_type=MediaType.HTML, include_in_schema=False)
    def swagger_ui(self, request: Request) -> Response:
        if not request.app.openapi_config:  # pragma: no cover
            raise ImproperlyConfigured(MSG_OPENAPI_NOT_INITIALIZED)

        if self.should_serve_endpoint(request):
            return Response(
                content=self.render_swagger_ui(request),
                status_code=HTTP_200_OK,
                media_type=MediaType.HTML,
            )
        return Response(
            content=self.render_404_page(),
            status_code=HTTP_404_NOT_FOUND,
            media_type=MediaType.HTML,
        )

    @get(path="/elements", media_type=MediaType.HTML, include_in_schema=False)
    def stoplight_elements(self, request: Request) -> Response:
        if not request.app.openapi_config:  # pragma: no cover
            raise ImproperlyConfigured(MSG_OPENAPI_NOT_INITIALIZED)

        if self.should_serve_endpoint(request):
            return Response(
                content=self.render_stoplight_elements(request),
                status_code=HTTP_200_OK,
                media_type=MediaType.HTML,
            )
        return Response(
            content=self.render_404_page(),
            status_code=HTTP_404_NOT_FOUND,
            media_type=MediaType.HTML,
        )

    @get(path="/redoc", media_type=MediaType.HTML, include_in_schema=False)
    def redoc(self, request: Request) -> Response:  # pragma: no cover
        if not request.app.openapi_config:  # pragma: no cover
            raise ImproperlyConfigured(MSG_OPENAPI_NOT_INITIALIZED)

        if self.should_serve_endpoint(request):
            return Response(
                content=self.render_redoc(request),
                status_code=HTTP_200_OK,
                media_type=MediaType.HTML,
            )
        return Response(
            content=self.render_404_page(),
            status_code=HTTP_404_NOT_FOUND,
            media_type=MediaType.HTML,
        )

    def render_swagger_ui(self, request: Request) -> str:
        schema = self.get_schema_from_request(request)
        # Note: Fix for Swagger rejection OpenAPI >=3.1
        if self._dumped_modified_schema == "":
            schema_copy = schema.copy()
            schema_copy.openapi = "3.0.3"
            self._dumped_modified_schema = dumps(
                schema_copy.json(by_alias=True, exclude_none=True),
                option=OPT_INDENT_2,
            ).decode("utf-8")

        head = f"""
          <head>
            <title>{schema.info.title}</title>
            {self.favicon}
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <link href="{self.swagger_css_url}" rel="stylesheet">
            <script src="{self.swagger_ui_bundle_js_url}" crossorigin></script>
            <script src="{self.swagger_ui_standalone_preset_js_url}" crossorigin></script>
            <style>{self.style}</style>
          </head>
        """
        body = f"""
          <body>
            <div id='swagger-container'/>
            <script type="text/javascript">
            const ui = SwaggerUIBundle({{
                spec: JSON.parse({self._dumped_modified_schema}),
                dom_id: '#swagger-container',
                deepLinking: true,
                showExtensions: true,
                showCommonExtensions: true,
                presets: [
                    SwaggerUIBundle.presets.apis,
                    SwaggerUIBundle.SwaggerUIStandalonePreset
                ],
            }})
            </script>
          </body>
        """
        return f"""
        <!DOCTYPE html>
            <html>
                {head}
                {body}
            </html>
        """

    def render_stoplight_elements(self, request: Request) -> str:
        schema = self.get_schema_from_request(request)
        head = f"""
          <head>
            <title>{schema.info.title}</title>
            {self.favicon}
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
            <link rel="stylesheet" href="{self.stoplight_elements_css_url}">
            <script src="{self.stoplight_elements_js_url}" crossorigin></script>
            <style>{self.style}</style>
          </head>
        """
        body = f"""
          <body>
            <elements-api
                apiDescriptionUrl="{self.path}/openapi.json"
                router="hash"
                layout="sidebar"
            />
          </body>
        """
        return f"""
        <!DOCTYPE html>
            <html>
                {head}
                {body}
            </html>
        """

    def render_redoc(self, request: Request) -> str:  # pragma: no cover
        schema = self.get_schema_from_request(request)
        if self._dumped_schema == "":
            self._dumped_schema = dumps(
                schema.json(by_alias=True, exclude_none=True),
                option=OPT_INDENT_2,
            ).decode("utf-8")
        head = f"""
          <head>
            <title>{schema.info.title}</title>
            {self.favicon}
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1">
            """
        if self.redoc_google_fonts:
            head += """
            <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
            """
        head += f"""
            <script src="{self.redoc_js_url}" crossorigin></script>
            <style>
                {self.style}
            </style>
          </head>
        """
        body = f"""
          <body>
            <div id='redoc-container'/>
            <script type="text/javascript">
                Redoc.init(
                    JSON.parse({self._dumped_schema}),
                    undefined,
                    document.getElementById('redoc-container')
                )
            </script>
          </body>
        """
        return f"""
        <!DOCTYPE html>
            <html>
                {head}
                {body}
            </html>
        """

    def render_404_page(self) -> str:
        """This method renders an HTML 404 page.

        Returns:
            A rendered html string.
        """

        return f"""
        <!DOCTYPE html>
        <html>
            <head>
                <title>404 Not found</title>
                {self.favicon}
                <meta charset="utf-8"/>
                <meta name="viewport" content="width=device-width, initial-scale=1">
                <style>
                    {self.style}
                </style>
            </head>
            <body>
                <h1>Error 404</h1>
            </body>
        </html>
        """



import hashlib

from esmerald.contrib.encoding import force_bytes


def pbkdf2(password, salt, iterations, dklen=0, digest=None):
    """Return the hash of password using pbkdf2."""
    if digest is None:
        digest = hashlib.sha256
    dklen = dklen or None
    password = force_bytes(password)
    salt = force_bytes(salt)
    return hashlib.pbkdf2_hmac(digest().name, password, salt, iterations, dklen)



"""
Helpers for the TortoiseORM integragion.
"""
from types import ModuleType
from typing import Dict, Iterable, List, Optional, Type, Union

from esmerald.conf import settings
from loguru import logger
from tortoise import Tortoise


async def init_database(
    config: Optional[dict] = None,
    config_file: Optional[str] = None,
    create_db: bool = False,
    db_url: Optional[str] = None,
    modules: Optional[Dict[str, Iterable[Union[str, ModuleType]]]] = None,
    use_tz: bool = False,
    timezone: str = "UTC",
    routers: Optional[List[Union[str, Type]]] = None,
) -> None:
    """
    Wrapper that establish a connection to the database.

    Example:
        from esmerald.contrib.tortoise import init_database
        from myapp.db import TORTOISE_ORM

        async def start_database():
            await init_database(config=TORTOISE_ORM)
    """
    logger.info("Starting database...")
    if not use_tz:
        use_tz = settings.use_tz
    if not timezone:
        timezone = settings.timezone
    await Tortoise.init(
        config=config,
        config_file=config_file,
        _create_db=create_db,
        db_url=db_url,
        modules=modules,
        use_tz=use_tz,
        timezone=timezone,
        routers=routers,
    )
    logger.info("Database started...")


async def stop_database() -> None:
    """
    Helper that stops the Tortoise Database.

    Example:
        from esmerald.contrib.tortoise import stop_database

        async def close_database():
            await stop_database()
    """
    logger.info("Closing connection to database")

    await Tortoise.close_connections()

    logger.info("Connection closed")



import traceback

from starlette.middleware.errors import (
    ServerErrorMiddleware as StarletteServerErrorMiddleware,
)
from starlette.responses import HTMLResponse, PlainTextResponse, Response

from esmerald.requests import Request


class ServerErrorMiddleware(StarletteServerErrorMiddleware):
    """
    Handles returning 500 responses when a server error occurs.

    If 'debug' is set, then traceback responses will be returned,
    otherwise the designated 'handler' will be called.

    This middleware class should generally be used to wrap *everything*
    else up, so that unhandled exceptions anywhere in the stack
    always result in an appropriate 500 response.
    """

    def generate_plain_text(self, exc: Exception) -> str:
        return "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))

    def debug_response(self, request: Request, exc: Exception) -> Response:
        accept = request.headers.get("accept", "")

        if "text/html" in accept:
            content = self.generate_html(exc)
            return HTMLResponse(content, status_code=500)
        content = self.generate_plain_text(exc)
        return PlainTextResponse(content, status_code=500)

    def error_response(self, request: Request, exc: Exception) -> Response:
        return PlainTextResponse("Internal Server Error", status_code=500)



from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Type,
    TypeVar,
    Union,
)

from pydantic.typing import AnyCallable
from starlette.middleware import Middleware as StarletteMiddleware  # noqa
from starlette.responses import Response as StarletteResponse
from starlette.types import ASGIApp
from typing_extensions import Literal

from esmerald.backgound import BackgroundTask, BackgroundTasks
from esmerald.exceptions import HTTPException, StarletteHTTPException
from esmerald.routing.gateways import WebSocketGateway
from esmerald.routing.router import Include
from esmerald.schedulers import AsyncIOScheduler

if TYPE_CHECKING:
    from esmerald.conf.global_settings import EsmeraldAPISettings  # noqa
    from esmerald.datastructures import Cookie, ResponseHeader, State  # noqa: TC004
    from esmerald.injector import Inject  # noqa
    from esmerald.protocols.middleware import MiddlewareProtocol
    from esmerald.requests import Request  # noqa
    from esmerald.responses import Response  # noqa
    from esmerald.routing.router import Gateway, HTTPHandler, Router  # noqa
    from esmerald.routing.views import APIView  # noqa
    from esmerald.websockets import WebSocket  # noqa
else:
    ASGIApp = Any
    HTTPHandler = Any
    Message = Any
    Receive = Any
    Scope = Any
    Send = Any
    StarletteMiddleware = Any
    BaseHTTPMiddleware = Any
    Inject = Any
    Request = Any
    WebSocket = Any
    State = Any
    Response = Any
    ResponseHeader = Any
    Cookie = Any
    Router = Any
    MiddlewareProtocol = Any
    APIView = Any
    Gateway = Any

AsyncAnyCallable = Callable[..., Awaitable[Any]]
HTTPMethod = Literal["GET", "POST", "DELETE", "PATCH", "PUT", "HEAD"]

Middleware = Union[
    StarletteMiddleware,
    Type[MiddlewareProtocol],
    Callable[..., ASGIApp],
]

ResponseType = Type[Response]

Dependencies = Dict[str, Inject]

CoroutineHandler = Coroutine[Any, Any, Response]
ExceptionHandler = Union[
    Callable[
        [Request, Union[Exception, HTTPException, StarletteHTTPException]],
        StarletteResponse,
    ],
    Callable[[Response, Any], CoroutineHandler],
]
ExceptionHandlers = Dict[Union[int, Type[Exception]], ExceptionHandler]

ReservedKwargs = Literal[
    "request",
    "socket",
    "headers",
    "query",
    "cookies",
    "state",
    "data",
]

ResponseHeaders = Dict[str, ResponseHeader]
ResponseCookies = List[Cookie]
AsyncAnyCallable = Callable[..., Awaitable[Any]]

LifeSpanHandler = Union[
    Callable[[], Any],
    Callable[[State], Any],
    Callable[[], Awaitable[Any]],
    Callable[[State], Awaitable[Any]],
]

SchedulerType = AsyncIOScheduler
DatetimeType = TypeVar("DatetimeType", bound=datetime)

ParentType = Union[APIView, Router]
APIGateHandler = Union[
    APIView,
    HTTPHandler,
    Router,
    AnyCallable,
    Gateway,
    WebSocketGateway,
]

RouteParent = Union["Router", "Include", "ASGIApp", "Gateway", "WebSocketGateway"]

BackgroundTaskType = Union[BackgroundTask, BackgroundTasks]
SecurityRequirement = Dict[str, List[str]]

ConnectionType = Union["Request", "WebSocket"]
DictStr = Dict[str, str]
DictAny = Dict[str, Any]
SettingsType = Type["EsmeraldAPISettings"]



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from esmerald.enums import ScopeType
from esmerald.protocols.middleware import MiddlewareProtocol
from pydantic import BaseConfig, BaseModel
from starlette.requests import HTTPConnection

if TYPE_CHECKING:
    from starlette.types import ASGIApp, Receive, Scope, Send


class AuthResult(BaseModel):
    user: Any

    class Config(BaseConfig):
        arbitrary_types_allowed = True


class BaseAuthMiddleware(ABC, MiddlewareProtocol):
    scopes = {ScopeType.HTTP, ScopeType.WEBSOCKET}

    def __init__(self, app: "ASGIApp"):
        super().__init__(app)
        self.app = app

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        auth_result = await self.authenticate(HTTPConnection(scope))
        scope["user"] = auth_result.user
        await self.app(scope, receive, send)

    @abstractmethod
    async def authenticate(self, request: HTTPConnection) -> AuthResult:  # pragma: no cover
        """
        The abstract method that needs to be implemented for any authentication middleware.
        """
        raise NotImplementedError("authenticate must be implemented.")



from typing import Any

from esmerald.utils.constants import IS_DEPENDENCY, SKIP_VALIDATION
from pydantic.fields import FieldInfo


def is_dependency_field(val: Any) -> bool:
    return bool(isinstance(val, FieldInfo) and bool(val.extra.get(IS_DEPENDENCY)))


def should_skip_dependency_validation(val: Any) -> bool:
    return bool(is_dependency_field(val) and val.extra.get(SKIP_VALIDATION))



from typing import Type, TypeVar, Union

from esmerald.permissions import BasePermission

PermissionType = TypeVar("PermissionType", bound=BasePermission)
Permission = Union[Type[PermissionType], BasePermission]



"""
Utils used by the Esmerald core management.
"""
import functools
import os
import pkgutil
from importlib import import_module
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from esmerald.core.management.typing import DirectiveType


def find_directives(management_dir):
    directive_dir = os.path.join(management_dir, "directives")
    directive_list = []
    for _, name, is_package in pkgutil.iter_modules([directive_dir]):
        if not is_package and not name.startswith("_"):
            directive_list.append(name)
    return directive_list


def load_directive_class(app_name: str, name: str) -> "DirectiveType":
    module = import_module("%s.management.directives.%s" % (app_name, name))
    return module.Directive()


@functools.lru_cache(maxsize=None)
def get_directives(location):
    command_list = find_directives(location)
    directives = {name: "esmerald.core" for name in command_list}

    return directives



from functools import partial
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Type, Union, cast

from openapi_schemas_pydantic import construct_open_api_with_schema_class
from openapi_schemas_pydantic.v3_1_0 import (
    Components,
    Contact,
    ExternalDocumentation,
    Info,
    License,
    OpenAPI,
    PathItem,
    Reference,
    SecurityRequirement,
    Server,
    Tag,
)
from openapi_schemas_pydantic.v3_1_0.path_item import PathItem
from pydantic import AnyUrl, BaseModel
from typing_extensions import Literal

from esmerald.enums import HttpMethod
from esmerald.openapi.path_item import create_path_item
from esmerald.routing.gateways import Gateway, WebSocketGateway
from esmerald.routing.router import Include
from esmerald.utils.helpers import is_class_and_subclass
from esmerald.utils.url import clean_path

if TYPE_CHECKING:
    from esmerald.applications import Esmerald
    from esmerald.openapi.apiview import OpenAPIView


class OpenAPIConfig(BaseModel):
    create_examples: bool = False
    openapi_apiview: Type["OpenAPIView"]
    title: str
    version: str
    contact: Optional[Contact] = None
    description: Optional[str] = None
    external_docs: Optional[ExternalDocumentation] = None
    license: Optional[License] = None
    security: Optional[List[SecurityRequirement]] = None
    components: Optional[Union[Components, List[Components]]] = None
    servers: List[Server] = [Server(url="/")]
    summary: Optional[str] = None
    tags: Optional[List[Tag]] = None
    terms_of_service: Optional[AnyUrl] = None
    use_handler_docstrings: bool = False
    webhooks: Optional[Dict[str, Union[PathItem, Reference]]] = None
    root_schema_site: Literal["redoc", "swagger", "elements"] = "redoc"
    enabled_endpoints: Set[str] = {
        "redoc",
        "swagger",
        "elements",
        "openapi.json",
        "openapi.yaml",
    }

    def to_openapi_schema(self) -> "OpenAPI":
        if isinstance(self.components, list):
            merged_components = Components()
            for components in self.components:
                for key in components.__fields__.keys():
                    value = getattr(components, key, None)
                    if value:
                        merged_value_dict = getattr(merged_components, key, {}) or {}
                        merged_value_dict.update(value)
                        setattr(merged_components, key, merged_value_dict)
            self.components = merged_components

        return OpenAPI(
            externalDocs=self.external_docs,
            security=self.security,
            components=cast("Components", self.components),
            servers=self.servers,
            tags=self.tags,
            webhooks=self.webhooks,
            info=Info(
                title=self.title,
                version=self.version,
                description=self.description,
                contact=self.contact,
                license=self.license,
                summary=self.summary,
                termsOfService=self.terms_of_service,
            ),
        )

    def get_http_verb(self, path_item):
        if getattr(path_item, "get", None):
            return HttpMethod.GET.value.lower()
        elif getattr(path_item, "post", None):
            return HttpMethod.POST.value.lower()
        elif getattr(path_item, "put", None):
            return HttpMethod.PUT.value.lower()
        elif getattr(path_item, "patch", None):
            return HttpMethod.PATCH.value.lower()
        elif getattr(path_item, "delete", None):
            return HttpMethod.DELETE.value.lower()
        elif getattr(path_item, "header", None):
            return HttpMethod.HEAD.value.lower()

    def create_openapi_schema_model(self, app: "Esmerald") -> "OpenAPI":
        from esmerald.applications import ChildEsmerald, Esmerald

        schema = self.to_openapi_schema()
        schema.paths = {}

        def parse_route(app, prefix=""):
            if not app.routes:
                return

            # Making sure that ChildEsmerald or esmerald
            if hasattr(app, "app"):
                if (
                    isinstance(app.app, (Esmerald, ChildEsmerald))
                    or (
                        is_class_and_subclass(app.app, Esmerald)
                        or is_class_and_subclass(app.app, ChildEsmerald)
                    )
                ) and not getattr(app.app, "enable_openapi", False):
                    return

            for route in app.routes:
                if isinstance(route, Include) and not route.include_in_schema:
                    continue

                if isinstance(route, WebSocketGateway):
                    continue

                if isinstance(route, Gateway):
                    if (
                        isinstance(route, Gateway)
                        and any(
                            handler.include_in_schema
                            for handler, _ in route.handler.route_map.values()
                        )
                        and (route.path_format or "/") not in schema.paths
                    ):
                        path = clean_path(prefix + route.path)
                        path_item = create_path_item(
                            route=route.handler,
                            create_examples=self.create_examples,
                            use_handler_docstrings=self.use_handler_docstrings,
                        )
                        verb = self.get_http_verb(path_item)
                        if not path in schema.paths:
                            schema.paths[path] = {}
                        if not verb in schema.paths[path]:
                            schema.paths[path][verb] = {}
                        schema.paths[path][verb] = getattr(path_item, verb, None)
                    continue

                route_app = getattr(route, "app", None)
                if not route_app:
                    continue

                if isinstance(route_app, partial):
                    try:
                        route_app = route_app.__wrapped__
                    except AttributeError:
                        pass

                path = clean_path(prefix + route.path)
                parse_route(route, prefix=f"{path}")

        parse_route(app)
        return construct_open_api_with_schema_class(schema)



from typing import Any, Dict

from orjson import OPT_OMIT_MICROSECONDS  # noqa
from orjson import OPT_SERIALIZE_NUMPY
from pydantic import BaseModel
from starlette.responses import JSONResponse as JSONResponse  # noqa

try:
    import orjson
except ImportError:  # pragma: nocover
    orjson = None  # type: ignore

try:
    import ujson
except ImportError:  # pragma: nocover
    ujson = None  # type: ignore


class BaseJSONResponse(JSONResponse):
    """
    Making sure it parses all the values from pydantic into dictionary.
    """

    @staticmethod
    def transform(value: Any) -> Dict[str, Any]:
        """
        Makes sure that every value is checked and if it's a pydantic model then parses into
        a dict().
        """
        if isinstance(value, BaseModel):
            return value.dict()
        raise TypeError("unsupported type")


class ORJSONResponse(BaseJSONResponse):
    def render(self, content: Any) -> bytes:
        assert orjson is not None, "orjson must be installed to use ORJSONResponse"
        return orjson.dumps(
            content,
            default=self.transform,
            option=OPT_SERIALIZE_NUMPY | OPT_OMIT_MICROSECONDS,
        )


class UJSONResponse(BaseJSONResponse):
    def render(self, content: Any) -> bytes:
        assert ujson is not None, "ujson must be installed to use UJSONResponse"
        return ujson.dumps(content, ensure_ascii=False, default=self.transform).encode("utf-8")



"""
Base classes for writing management directives (named directives which can
be rund through `esmerald-admin`).
"""
import argparse
import os
import sys
from typing import Any, Dict

import esmerald
from esmerald.core.management.exceptions import DirectiveError, SystemCheckError
from esmerald.core.management.formatters import EsmeraldHelpFormatter, OutputWrapper
from esmerald.core.management.parsers import DirectiveParser


class BaseDirective:
    """
    The base class from which all management directives derive.
    """

    help = ""

    _called_from_command_line = False
    base_stealth_options = ("stderr", "stdout")
    stealth_options = ()
    suppressed_base_arguments = set()

    def __init__(self, stdout=None, stderr=None):
        self.stdout = OutputWrapper(stdout or sys.stdout)
        self.stderr = OutputWrapper(stderr or sys.stderr)

    def get_version(self):
        """
        Returns the current version of Esmerald.
        """
        return esmerald.__version__

    def create_parser(self, prog_name: Any, subdirective: Any, **kwargs: Dict[str, Any]):
        """
        Create and return the ``ArgumentParser`` which will be used to
        parse the arguments to this directive.
        """
        kwargs.setdefault("formatter_class", EsmeraldHelpFormatter)
        parser = DirectiveParser(
            prog="%s %s" % (os.path.basename(prog_name), subdirective),
            description=self.help or None,
            missing_args_message=getattr(self, "missing_args_message", None),
            called_from_command_line=getattr(self, "_called_from_command_line", None),
            **kwargs,
        )
        self.add_base_argument(
            parser,
            "--version",
            action="version",
            version=self.get_version(),
            help="Show program's version number and exit.",
        )
        self.add_base_argument(
            parser,
            "-v",
            "--verbosity",
            default=1,
            type=int,
            choices=[0, 1, 2, 3],
            help=(
                "Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, "
                "3=very verbose output"
            ),
        )
        self.add_base_argument(
            parser,
            "--traceback",
            action="store_true",
            help="Raise on DirectiveError exceptions.",
        )
        self.add_arguments(parser)
        return parser

    def add_arguments(self, parser: Any):
        """
        Entrypoint for directives to add custom arguments.
        """

    def add_base_argument(self, parser: Any, *args: Any, **kwargs: Dict[str, Any]):
        for arg in args:
            if arg in self.suppressed_base_arguments:
                kwargs["help"] = argparse.SUPPRESS
                break
        parser.add_argument(*args, **kwargs)

    def print_help(self, prog_name, subdirective):
        parser = self.create_parser(prog_name, subdirective)
        parser.print_help()

    def run_from_argv(self, argv: Any):
        self._called_from_command_line = True
        parser = self.create_parser(argv[0], argv[1])

        options = parser.parse_args(argv[2:])
        cmd_options = vars(options)
        # Move positional args out of options to mimic legacy optparse
        args = cmd_options.pop("args", ())
        try:
            self.run(*args, **cmd_options)
        except DirectiveError as e:
            if options.traceback:
                raise
            if isinstance(e, SystemCheckError):
                self.stderr.write(str(e), lambda x: x)
            else:
                self.stderr.write("%s: %s" % (e.__class__.__name__, e))
            sys.exit(e.returncode)

    def run(self, *args: Any, **options: Dict[str, Any]):
        if options.get("stdout"):
            self.stdout = OutputWrapper(options["stdout"])
        if options.get("stderr"):
            self.stderr = OutputWrapper(options["stderr"])

        output = self.handle(*args, **options)
        if output:
            self.stdout.write(output)
        return output

    def handle(self, *args: Any, **options: Dict[str, Any]):
        """
        The logic of the directive. Subclasses must implement this method.
        """
        raise NotImplementedError("subclasses of BaseDirective must provide a handle() method")



from typing import Any, Callable, Optional

from esmerald import injector


def Injects(  # noqa: N802
    dependency: Optional[Callable[..., Any]] = None, *, use_cache: bool = True
) -> Any:
    return injector.Inject(dependency=dependency, use_cache=use_cache)



from typing import Type

from esmerald.enums import MediaType
from pydantic import BaseModel


class ResponseSpecification(BaseModel):
    model: Type[BaseModel]
    create_examples: bool = True
    description: str = "Additional response"
    media_type: MediaType = MediaType.JSON



from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from pydantic.fields import ModelField


def is_field_optional(field: "ModelField") -> bool:
    """
    Returns bool True or False for the optional model field.
    """
    return not field.required and not is_any_type(field=field) and field.allow_none


def is_any_type(field: "ModelField") -> bool:
    """
    Checks if the field is of type Any.
    """
    name = cast("Any", getattr(field.outer_type_, "_name", None))
    return (name is not None and "Any" in name) or field.type_ is Any



from typing import TYPE_CHECKING, Any, Dict, List, cast

from esmerald.enums import ParamType
from esmerald.exceptions import ImproperlyConfigured
from esmerald.openapi.schema import create_schema
from esmerald.utils.constants import REQUIRED, RESERVED_KWARGS
from esmerald.utils.dependency import is_dependency_field
from openapi_schemas_pydantic.v3_1_0.parameter import Parameter
from pydantic.fields import Undefined


if TYPE_CHECKING:
    from esmerald.routing.router import HTTPHandler
    from esmerald.types import Dependencies
    from openapi_schemas_pydantic.v3_1_0.schema import Schema
    from pydantic import BaseModel
    from pydantic.fields import ModelField


def create_path_parameter_schema(
    path_parameter: Any, field: "ModelField", create_examples: bool
) -> "Schema":
    field.sub_fields = None
    field.outer_type_ = path_parameter["type"]
    return create_schema(field=field, create_examples=create_examples)


class ParameterCollection:
    def __init__(self, handler: "HTTPHandler") -> None:
        self.handler = handler
        self._parameters: Dict[str, Parameter] = {}

    def add(self, parameter: Parameter) -> None:
        if parameter.name not in self._parameters:
            self._parameters[parameter.name] = parameter
            return
        pre_existing = self._parameters[parameter.name]
        if parameter == pre_existing:
            return
        raise ImproperlyConfigured(
            f"OpenAPI schema generation for handler `{self.handler}` detected multiple parameters named "
            f"'{parameter.name}' with different types."
        )

    def list(self) -> List[Parameter]:
        return list(self._parameters.values())


def create_parameter(
    model_field: "ModelField",
    parameter_name: str,
    path_paramaters: Any,
    create_examples: bool,
) -> Parameter:
    schema = None
    is_required = (
        cast("bool", model_field.required) if model_field.required is not Undefined else False
    )
    extra = model_field.field_info.extra
    if any(path_param["name"] == parameter_name for path_param in path_paramaters):
        param_in = ParamType.PATH
        is_required = True
        path_parameter = [p for p in path_paramaters if parameter_name in p["name"]][0]
        schema = create_path_parameter_schema(
            path_parameter=path_parameter,
            field=model_field,
            create_examples=create_examples,
        )
    elif extra.get(ParamType.HEADER):
        parameter_name = extra[ParamType.HEADER]
        param_in = ParamType.HEADER
        is_required = model_field.field_info.extra[REQUIRED]
    elif extra.get(ParamType.COOKIE):
        parameter_name = extra[ParamType.COOKIE]
        param_in = ParamType.COOKIE
        is_required = model_field.field_info.extra[REQUIRED]
    else:
        param_in = ParamType.QUERY
        parameter_name = extra.get(ParamType.QUERY) or parameter_name

    if not schema:
        schema = create_schema(field=model_field, create_examples=create_examples)

    return Parameter(
        name=parameter_name,
        param_in=param_in,
        required=is_required,
        param_schema=schema,
        description=schema.description,
    )


def get_recursive_handler_parameters(
    field_name: str,
    model_field: "ModelField",
    dependencies: "Dependencies",
    handler: "HTTPHandler",
    path_parameters: Any,
    create_examples: bool,
):
    if field_name not in dependencies:
        return [
            create_parameter(
                model_field=model_field,
                parameter_name=field_name,
                path_paramaters=path_parameters,
                create_examples=create_examples,
            )
        ]
    dependency_fields = cast("BaseModel", dependencies[field_name].signature_model).__fields__

    return create_parameter_for_handler(
        handler, dependency_fields, path_parameters, create_examples
    )


def create_parameter_for_handler(
    handler: "HTTPHandler",
    handler_fields: Dict[str, "ModelField"],
    path_parameters: Any,
    create_examples: bool,
):

    parameters = ParameterCollection(handler=handler)
    dependencies = handler.get_dependencies()

    for field_name, model_field in filter(
        lambda items: items[0] not in RESERVED_KWARGS and items[0] not in {},
        handler_fields.items(),
    ):
        if is_dependency_field(model_field.field_info) and field_name not in dependencies:
            continue

        for parameter in get_recursive_handler_parameters(
            field_name=field_name,
            model_field=model_field,
            dependencies=dependencies,
            handler=handler,
            path_parameters=path_parameters,
            create_examples=create_examples,
        ):
            parameters.add(parameter)

    return parameters.list()



from .base import (
    AllowAny,
    BasePermission,
    DenyAll,
    IsAdminUser,
    IsAuthenticated,
    IsAuthenticatedOrReadOnly,
)

__all__ = [
    "AllowAny",
    "BasePermission",
    "DenyAll",
    "IsAdminUser",
    "IsAuthenticated",
    "IsAuthenticatedOrReadOnly",
]



from typing import TYPE_CHECKING, Any, List, TypeVar

from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from esmerald.types import DictAny

T = TypeVar("T")


@runtime_checkable
class AsyncDAOProtocol(Protocol):  # pragma: no cover
    """
    Async Data Access Object (AsyncDAO) is an abstract pattern widely used to
    separate the underlying logic of the database from the business logic.

    Esmerald definition of an AsyncDAO will contain the base patterns for an object.
    """

    model: Any

    """
    The model object. Usually this object is what allows the connection and
    data access to a database.
    """

    async def get(self, obj_id: Any, **kwargs: "DictAny") -> Any:
        ...

    async def get_all(self, **kwargs: "DictAny") -> List[Any]:
        ...

    async def update(self, obj_id: Any, **kwargs: "DictAny") -> Any:
        ...

    async def delete(self, obj_id: Any, **kwargs: "DictAny") -> Any:
        ...

    async def create(self, **kwargs: "DictAny") -> Any:
        ...



from typing import List, Optional

from pydantic import BaseModel


class CORSConfig(BaseModel):
    allow_origins: List[str] = ["*"]
    allow_methods: List[str] = ["*"]
    allow_headers: List[str] = ["*"]
    allow_credentials: bool = False
    allow_origin_regex: Optional[str] = None
    expose_headers: List[str] = []
    max_age: int = 600



from typing import TYPE_CHECKING, Any, List, TypeVar

from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from esmerald.types import DictAny

T = TypeVar("T")


@runtime_checkable
class DaoProtocol(Protocol):  # pragma: no cover
    """
    Data Access Object (DAO) is an abstract pattern widely used to
    separate the underlying logic of the database from the business logic.

    Esmerald definition of a DAO will contain the base patterns for an object
    """

    @property
    def model(self) -> Any:
        ...

    """
    The model object. Usually this object is what allows the connection and
    data access to a database.
    """

    def get(self, obj_id: Any, **kwargs: "DictAny") -> Any:
        ...

    def get_all(self, **kwargs: "DictAny") -> List[Any]:
        ...

    def update(self, obj_id: Any, **kwargs: "DictAny") -> Any:
        ...

    def delete(self, obj_id: Any, **kwargs: "DictAny") -> Any:
        ...

    def create(self, **kwargs: "DictAny") -> Any:
        ...



from typing import Type, TypeVar, Union

from esmerald.core.management.base import BaseDirective

DiretiveTypeVar = TypeVar("DiretiveTypeVar", bound=BaseDirective)
DirectiveType = Union[Type[DiretiveTypeVar], BaseDirective]



from argparse import ArgumentParser
from typing import Any, Dict, Optional

from esmerald.core.management.exceptions import DirectiveError


class DirectiveParser(ArgumentParser):
    """
    Customized ArgumentParser class to improve some error messages and prevent SystemExit.
    """

    def __init__(
        self,
        *,
        missing_args_message: Optional[str] = None,
        called_from_command_line: Optional[Any] = None,
        **kwargs: Dict[str, Any]
    ):
        self.missing_args_message = missing_args_message
        self.called_from_command_line = called_from_command_line
        super().__init__(**kwargs)

    def parse_args(self, args: Any = None, namespace: str = None):
        if self.missing_args_message and not (
            args or any(not arg.startswith("-") for arg in args)
        ):
            self.error(self.missing_args_message)
        return super().parse_args(args, namespace)

    def error(self, message):
        if self.called_from_command_line:
            super().error(message)
        else:
            raise DirectiveError("Error: %s" % message)



from contextlib import suppress
from typing import TYPE_CHECKING, Any

from orjson import JSONDecodeError, loads
from pydantic import BaseConfig, BaseModel
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON

from esmerald.datastructures import UploadFile
from esmerald.enums import EncodingType

if TYPE_CHECKING:
    from pydantic.fields import ModelField
    from pydantic.typing import DictAny
    from starlette.datastructures import FormData


class HashableBaseModel(BaseModel):
    """
    Pydantic BaseModel by default doesn't handle with hashable types the same way
    a python object would and therefore there are types that are mutable (list, set)
    not hashable and those need to be handled properly.

    HashableBaseModel handles those corner cases.
    """

    def __hash__(self):
        values = {}
        for key, value in self.__dict__.items():
            values[key] = None
            if isinstance(value, (list, set)):
                values[key] = tuple(value)
            else:
                values[key] = value
        return hash((type(self),) + tuple(values))


class ArbitraryHashableBaseModel(HashableBaseModel):
    """
    Same as HashableBaseModel but allowing arbitrary values
    """

    class Config(HashableBaseModel.Config):
        extra = "allow"
        arbitrary_types_allowed = True


class BaseModelExtra(BaseModel):
    """
    BaseModel that allows extra to be passed.
    """

    class Config:
        extra = "allow"


class ArbitraryBaseModel(BaseModel):
    """
    ArbitratyBaseModel that allows arbitrary_types_allowed to be passed.
    """

    class Config(BaseConfig):
        arbitrary_types_allowed = True


def validate_media_type(field: "ModelField", values: "DictAny"):
    """
    Validates the media type against the available types.
    """
    if field.shape in SHAPE_LIST:
        return list(values.values())
    if field.shape in SHAPE_SINGLETON and field.type_in[UploadFile] and values:
        return list(values.values())[0]


def parse_form_data(media_type: "EncodingType", form_data: "FormData", field: "ModelField") -> Any:
    values: "DictAny" = {}
    for key, value in form_data.multi_items():
        if not isinstance(value, UploadFile):
            with suppress(JSONDecodeError):
                value = loads(value)
        if isinstance(value, UploadFile):
            value = UploadFile(
                filename=value.filename,
                file=value.file,
                content_type=value.content_type,
                headers=value.headers,
            )
        if values.get(key):
            if isinstance(values[key], list):
                values[key].append(value)
            else:
                values[key] = [values[key], value]
        else:
            values[key] = value
    if media_type == EncodingType.MULTI_PART:
        return validate_media_type(field=field, values=values)
    return values



from .base import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    Response,
    StarletteResponse,
    StreamingResponse,
)
from .json import ORJSONResponse, UJSONResponse
from .template import TemplateResponse

__all__ = [
    "FileResponse",
    "HTMLResponse",
    "JSONResponse",
    "ORJSONResponse",
    "PlainTextResponse",
    "Response",
    "StarletteResponse",
    "StreamingResponse",
    "TemplateResponse",
    "UJSONResponse",
]



from typing import TYPE_CHECKING, Any, Dict, Generic, NoReturn, Optional, TypeVar, Union, cast

from orjson import OPT_OMIT_MICROSECONDS  # noqa
from orjson import OPT_SERIALIZE_NUMPY, dumps
from pydantic import BaseModel
from starlette import status
from starlette.responses import FileResponse as FileResponse  # noqa
from starlette.responses import HTMLResponse as HTMLResponse  # noqa
from starlette.responses import JSONResponse as JSONResponse  # noqa
from starlette.responses import PlainTextResponse as PlainTextResponse  # noqa
from starlette.responses import RedirectResponse as RedirectResponse  # noqa
from starlette.responses import Response as StarletteResponse  # noqa
from starlette.responses import StreamingResponse as StreamingResponse  # noqa

from esmerald.enums import MediaType
from esmerald.exceptions import ImproperlyConfigured

if TYPE_CHECKING:
    from esmerald.backgound import BackgroundTask, BackgroundTasks
    from esmerald.types import ResponseCookies

T = TypeVar("T")


class Response(StarletteResponse, Generic[T]):
    def __init__(
        self,
        content: T,
        *,
        status_code: Optional[int] = status.HTTP_200_OK,
        media_type: Optional[Union["MediaType", str]] = MediaType.JSON,
        background: Optional[Union["BackgroundTask", "BackgroundTasks"]] = None,
        headers: Optional[Dict[str, Any]] = None,
        cookies: Optional["ResponseCookies"] = None,
    ) -> None:
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers or {},
            media_type=media_type,
            background=cast("BackgroundTask", background),
        )
        self.cookies = cookies or []

    @staticmethod
    def transform(value: Any) -> Dict[str, Any]:
        if isinstance(value, BaseModel):
            return value.dict()
        raise TypeError("unsupported type")  # pragma: no cover

    def render(self, content: Any) -> bytes:
        try:
            if (
                content is None
                or content is NoReturn
                and (
                    self.status_code < 100
                    or self.status_code
                    in {status.HTTP_204_NO_CONTENT, status.HTTP_304_NOT_MODIFIED}
                )
            ):
                return b""
            if self.media_type == MediaType.JSON:
                return dumps(
                    content,
                    default=self.transform,
                    option=OPT_SERIALIZE_NUMPY | OPT_OMIT_MICROSECONDS,
                )
            return super().render(content)
        except (AttributeError, ValueError, TypeError) as e:
            raise ImproperlyConfigured("Unable to serialize response content") from e



from typing import List, Type, Union

from esmerald.protocols.template import TemplateEngineProtocol
from esmerald.template.jinja import JinjaTemplateEngine
from pydantic import BaseConfig, BaseModel, DirectoryPath


class TemplateConfig(BaseModel):
    class Config(BaseConfig):
        arbitrary_types_allowed = True

    directory: Union[DirectoryPath, List[DirectoryPath]]
    engine: Type[TemplateEngineProtocol] = JinjaTemplateEngine



"""
Timezone-related classes and functions.
"""
import functools
import sys
from typing import Any, Optional, Union
from zoneinfo import ZoneInfo

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

from datetime import date, datetime, timedelta, timezone

from asgiref.local import Local
from esmerald.conf import settings

__all__ = [
    "utc",
    "get_fixed_timezone",
    "get_default_timezone",
    "get_default_timezone_name",
    "get_current_timezone",
    "get_current_timezone_name",
    "localtime",
    "localdate",
    "now",
    "is_aware",
    "is_naive",
    "make_aware",
    "make_naive",
]

NOT_PASSED = object()


def __getattr__(name: str) -> timezone.utc:
    if name != "utc":
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    return timezone.utc


def get_fixed_timezone(offset: Any) -> timezone:
    """Return a tzinfo instance with a fixed offset from UTC."""
    if isinstance(offset, timedelta):
        offset = offset.total_seconds() // 60
    sign = "-" if offset < 0 else "+"
    hhmm = "%02d%02d" % divmod(abs(offset), 60)
    name = sign + hhmm
    return timezone(timedelta(minutes=offset), name)


# In order to avoid accessing settings at compile time,
# wrap the logic in a function and cache the result.


@functools.lru_cache
def get_default_timezone() -> ZoneInfo:
    """
    Return the default time zone as a tzinfo instance.

    This is the time zone defined by settings.timezone.
    """
    return zoneinfo.ZoneInfo(settings.timezone)


# This function exists for consistency with get_current_timezone_name
def get_default_timezone_name() -> str:
    """Return the name of the default time zone."""
    return _get_timezone_name(get_default_timezone())


_active = Local()


def get_current_timezone() -> ZoneInfo:
    """Return the currently active time zone as a tzinfo instance."""
    return getattr(_active, "value", get_default_timezone())


def get_current_timezone_name() -> str:
    """Return the name of the currently active time zone."""
    return _get_timezone_name(get_current_timezone())


def _get_timezone_name(timezone: timezone) -> Union[timezone, str]:
    """
    Return the offset for fixed offset timezones, or the name of timezone if
    not set.
    """
    return timezone.tzname(None) or str(timezone)


# Utilities


def localtime(value=None, timezone=None) -> Union[datetime, Any]:
    """
    Convert an aware datetime.datetime to local time.

    Only aware datetimes are allowed. When value is omitted, it defaults to
    now().

    Local time is defined by the current time zone, unless another time zone
    is specified.
    """
    if value is None:
        value = now()
    if timezone is None:
        timezone = get_current_timezone()
    # Emulate the behavior of astimezone() on Python < 3.6.
    if is_naive(value):
        raise ValueError("localtime() cannot be applied to a naive datetime")
    return value.astimezone(timezone)


def localdate(value=None, timezone=None) -> date:
    """
    Convert an aware datetime to local time and return the value's date.

    Only aware datetimes are allowed. When value is omitted, it defaults to
    now().

    Local time is defined by the current time zone, unless another time zone is
    specified.
    """
    return localtime(value, timezone).date()


def now() -> datetime:
    """
    Return an aware or naive datetime.datetime, depending on settings.use_tz.
    """
    return datetime.now(tz=timezone.utc if settings.use_tz else None)


# By design, these four functions don't perform any checks on their arguments.
# The caller should ensure that they don't receive an invalid value like None.


def is_aware(value: datetime) -> Union[timedelta, None]:
    """
    Determine if a given datetime.datetime is aware.

    The concept is defined in Python's docs:
    https://docs.python.org/library/datetime.html#datetime.tzinfo

    Assuming value.tzinfo is either None or a proper datetime.tzinfo,
    value.utcoffset() implements the appropriate logic.
    """
    return value.utcoffset() is not None


def is_naive(value: datetime) -> Union[timedelta, None]:
    """
    Determine if a given datetime.datetime is naive.

    The concept is defined in Python's docs:
    https://docs.python.org/library/datetime.html#datetime.tzinfo

    Assuming value.tzinfo is either None or a proper datetime.tzinfo,
    value.utcoffset() implements the appropriate logic.
    """
    return value.utcoffset() is None


def make_aware(value, timezone: Optional[timezone] = None, is_dst=NOT_PASSED):
    """Make a naive datetime.datetime in a given time zone aware."""
    if is_dst is NOT_PASSED:
        is_dst = None

    if timezone is None:
        timezone = get_current_timezone()
    if _is_pytz_zone(timezone):
        # This method is available for pytz time zones.
        return timezone.localize(value, is_dst=is_dst)
    else:
        # Check that we won't overwrite the timezone of an aware datetime.
        if is_aware(value):
            raise ValueError("make_aware expects a naive datetime, got %s" % value)
        # This may be wrong around DST changes!
        return value.replace(tzinfo=timezone)


def make_naive(value, timezone=None):
    """Make an aware datetime.datetime naive in a given time zone."""
    if timezone is None:
        timezone = get_current_timezone()
    # Emulate the behavior of astimezone() on Python < 3.6.
    if is_naive(value):
        raise ValueError("make_naive() cannot be applied to a naive datetime")
    return value.astimezone(timezone).replace(tzinfo=None)


_PYTZ_IMPORTED = False


def _pytz_imported() -> bool:
    """
    Detects whether or not pytz has been imported without importing pytz.

    Copied from pytz_deprecation_shim with thanks to Paul Ganssle.
    """
    global _PYTZ_IMPORTED

    if not _PYTZ_IMPORTED and "pytz" in sys.modules:
        _PYTZ_IMPORTED = True

    return _PYTZ_IMPORTED


def _is_pytz_zone(tz) -> Any:
    """Checks if a zone is a pytz zone."""
    # See if pytz was already imported rather than checking
    # settings.USE_DEPRECATED_PYTZ to *allow* manually passing a pytz timezone,
    # which some of the test cases (at least) rely on.
    if not _pytz_imported():
        return False

    # If tz could be pytz, then pytz is needed here.
    import pytz

    _PYTZ_BASE_CLASSES = (pytz.tzinfo.BaseTzInfo, pytz._FixedOffset)
    # In releases prior to 2018.4, pytz.UTC was not a subclass of BaseTzInfo
    if not isinstance(pytz.UTC, pytz._FixedOffset):
        _PYTZ_BASE_CLASSES = _PYTZ_BASE_CLASSES + (type(pytz.UTC),)

    return isinstance(tz, _PYTZ_BASE_CLASSES)


_DIR = dir()


def __dir__():
    return sorted([*_DIR, "utc"])



from esmerald.enums import HttpMethod
from starlette import status

RESERVED_KWARGS = {"state", "headers", "cookies", "request", "socket", "data", "query"}
REQUIRED = "required"
IS_DEPENDENCY = "is_dependency"
SKIP_VALIDATION = "skip_validation"

REDIRECT_STATUS_CODES = {
    status.HTTP_301_MOVED_PERMANENTLY,
    status.HTTP_302_FOUND,
    status.HTTP_303_SEE_OTHER,
    status.HTTP_307_TEMPORARY_REDIRECT,
    status.HTTP_308_PERMANENT_REDIRECT,
}

SOCKET = "socket"
DATA = "data"
REQUEST = "request"

AVAILABLE_METHODS = [
    HttpMethod.GET,
    HttpMethod.HEAD,
    HttpMethod.PATCH,
    HttpMethod.PUT,
    HttpMethod.POST,
    HttpMethod.DELETE,
    HttpMethod.OPTIONS,
]
